# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import asyncio
from collections.abc import Sequence

from packaging.utils import canonicalize_name
from packaging.version import Version

from .. import errors, model
from .core import SimpleRepository


class PrioritySelectedProjectsRepository(SimpleRepository):
    """Group together multiple repositories, using a first-seen policy

    When the list of all projects is requested, returns the union of all
    projects in all sources. When a specific project is requested, returns
    the first source which has the project, and does not include project
    information from any other source. Note that the project name being
    requested is normalized, to absolutely prevent against dependency
    confusion attacks from sources later in the sequence.
    """
    def __init__(self, sources: Sequence[SimpleRepository]) -> None:
        if len(sources) < 2:
            raise ValueError(
                "A priority selected repository must have two or more "
                "source repositories",
            )
        self.sources = sources

    async def get_project_page(
        self,
        project_name: str,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.ProjectDetail:
        """Retrieves a project page for the specified normalized project name
        by searching through the grouped list of sources in a first seen policy.
        """
        for source in self.sources:
            try:
                project_page = await source.get_project_page(
                    project_name,
                    request_context=request_context,
                )
            except errors.PackageNotFoundError:
                continue
            return project_page

        raise errors.PackageNotFoundError(
            package_name=project_name,
        )

    async def get_project_list(
        self,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.ProjectList:
        """Retrieves a combined list of projects from all the sources."""
        results: list[model.ProjectList | BaseException] = await asyncio.gather(
            *(
                source.get_project_list(request_context=request_context)
                for source in self.sources
            ),
            return_exceptions=True,
        )

        project_lists = [item for item in results if isinstance(item, model.ProjectList)]

        if len(project_lists) != len(self.sources):
            # TODO: Use an exception group to raise
            # multiple exceptions together.
            any_exception = next(item for item in results if isinstance(item, BaseException))
            raise any_exception

        projects = set().union(
            *(
                index.projects for index in project_lists
            ),
        )

        # Downgrade the API version to the lowest available, as it will not be
        # possible to calculate the missing files to perform a version upgrade.
        api_version = min(Version(project.meta.api_version) for project in project_lists)
        return model.ProjectList(
            meta=model.Meta(str(api_version)),
            projects=frozenset(
                model.ProjectListElement(
                    name=canonicalize_name(p.name),
                ) for p in projects
            ),
        )

    async def get_resource(
        self,
        project_name: str,
        resource_name: str,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.Resource:
        for source in self.sources:
            try:
                resource = await source.get_resource(
                    project_name,
                    resource_name,
                    request_context=request_context,
                )
            except errors.ResourceUnavailable:
                pass
            else:
                return resource
        raise errors.ResourceUnavailable(resource_name)



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from packaging.version import Version
import pytest

from .. import packaging


@pytest.mark.parametrize(
        ("filename", "package_version"), [
            ("my_package1-0.0.1-any.whl", "0.0.1"),
            ("my-package1-0.0.1.tar.gz", "0.0.1"),
            ("my-package1-0.10.zip", "0.10"),
        ],
)
def test_extract_package_version(filename: str, package_version: str | None) -> None:
    assert packaging.extract_package_version(filename, "my-package1") == package_version


def test_extract_package_version_failed() -> None:
    with pytest.raises(
        ValueError,
        match="hello-0.1 does not match non_matching",
    ):
        packaging.extract_package_version("hello-0.1.zip", "non_matching")


@pytest.mark.parametrize(
        ("filename", "package_format"), [
            ("my_package-0.0.1-any.whl", "wheel"),
            ("my-package-0.0.1.tar.gz", "sdist"),
            ("my-package-0.0.1.jpeg", "other format"),
        ],
)
def test_extract_package_format(filename: str, package_format: str) -> None:
    assert packaging.extract_package_format(filename).value == package_format


@pytest.mark.parametrize(
    "filename, project_name, version", [
        ("numpy-1.6.0-cp26-cp26m-manylinux1_x86_64.whl", "numpy", Version("1.6.0")),
        ("numpy-1.6.0.tar.gz", "numpy", Version("1.6.0")),
        ("numpy-1.6.0.tar.gz", "tensorflow", Version("0.0rc0")),
        ("numpy-aaaa.whl", "numpy", Version("0.0rc0")),
        ("numpy-aaaa.tar.gz", "numpy", Version("0.0rc0")),
    ],
)
def test__safe_version(
    filename: str,
    project_name: str,
    version: Version,
) -> None:
    assert packaging.safe_version(filename, project_name) == version



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from datetime import datetime
import os
from pathlib import Path

import pytest

from ... import errors, model
from ...components.local import LocalRepository


def test_path_resolution(tmp_path: Path) -> None:
    path = tmp_path / "my-path" / "simple"
    path.mkdir(parents=True)
    symlink = Path(tmp_path / "symlink")
    symlink.symlink_to(path)
    repo = LocalRepository(symlink)
    assert repo._index_path == tmp_path / "my-path" / "simple"


@pytest.fixture
def simple_dir(tmp_path: Path) -> Path:
    for project in ("numpy", "tensorflow", "pandas", ".not_normalized"):
        (tmp_path / project).mkdir()
    for file in ("numpy-1.0-any.whl", "numpy-1.1.tar.gz"):
        (tmp_path / "numpy" / file).write_text("content")
    (tmp_path / ".not_normalized" / "unreachable_resource").touch()
    return tmp_path


@pytest.fixture
def repository(simple_dir: Path) -> LocalRepository:
    return LocalRepository(simple_dir)


@pytest.mark.asyncio
async def test_get_project_list(simple_dir: Path) -> None:
    repo = LocalRepository(
        index_path=simple_dir,
    )
    project_list = await repo.get_project_list()

    assert project_list == model.ProjectList(
        meta=model.Meta("1.0"),
        projects=frozenset([
            model.ProjectListElement("numpy"),
            model.ProjectListElement("tensorflow"),
            model.ProjectListElement("pandas"),
        ]),
    )


@pytest.mark.asyncio
async def test_get_resource(simple_dir: Path) -> None:
    repo = LocalRepository(
        index_path=simple_dir,
    )

    resource_path = simple_dir / "numpy" / "numpy-1.0-any.whl"
    resource_path.touch()
    time = 946940400.0
    os.utime(resource_path, (time, time))
    # Expected etag given the mtime and size
    etag = "fc4e65c49baf52fa6e8fa52d539a153e"

    resource = await repo.get_resource("numpy", "numpy-1.0-any.whl")

    assert resource == model.LocalResource(
        path=resource_path,
        to_cache=False,
        context=model.Context(etag=etag),
    )


@pytest.mark.parametrize(
    ("project, resource"), [
        ("numpy", "numpy-2.0.tar.gz"),
        ("seaborn", "seaborn-1.0.tar.gz"),
    ],
)
@pytest.mark.asyncio
async def test_get_resource__unavailable(
    simple_dir: Path,
    project: str,
    resource: str,
) -> None:
    repo = LocalRepository(
        index_path=simple_dir,
    )

    with pytest.raises(
        errors.ResourceUnavailable,
        match=f"Resource '{resource}' was not found in the configured source",
    ):
        await repo.get_resource(project, resource)


@pytest.mark.parametrize(
    ("project, resource"), [
        ("numpy", "../../../etc/password"),
        ("tensorflow", "../numpy/numpy-1.0.tar.gz"),
    ],
)
@pytest.mark.asyncio
async def test_get_resource__path_traversal(
    simple_dir: Path,
    project: str,
    resource: str,
) -> None:
    repo = LocalRepository(
        index_path=simple_dir,
    )
    with pytest.raises(
        ValueError,
        match=f"{(simple_dir / project /resource).resolve()} is not contained in {repo._index_path / project}",
    ):
        await repo.get_resource(project, resource)


@pytest.mark.asyncio
async def test_get_project_page(simple_dir: Path) -> None:
    repo = LocalRepository(
        index_path=simple_dir,
    )
    project_details = await repo.get_project_page("numpy")
    assert project_details == model.ProjectDetail(
        meta=model.Meta("1.1"),
        name="numpy",
        files=(
            model.File(
                filename='numpy-1.0-any.whl',
                url="file://" + str(simple_dir / 'numpy/numpy-1.0-any.whl'),
                # On the fly hashes currently disabled for local repository.
                hashes={
                    # "sha256": sha256sum(simple_dir / "numpy" / "numpy-1.0-any.whl"),
                },
                upload_time=datetime.utcfromtimestamp(
                        os.path.getmtime(simple_dir / "numpy" / "numpy-1.0-any.whl"),
                ),
                size=os.stat(simple_dir / "numpy" / "numpy-1.0-any.whl").st_size,
            ),
            model.File(
                filename='numpy-1.1.tar.gz',
                url="file://" + str(simple_dir / 'numpy/numpy-1.1.tar.gz'),
                hashes={
                    # "sha256": sha256sum(simple_dir / "numpy" / "numpy-1.1.tar.gz"),
                },
                upload_time=datetime.utcfromtimestamp(
                        os.path.getmtime(simple_dir / "numpy" / "numpy-1.1.tar.gz"),
                ),
                size=os.stat(simple_dir / "numpy" / "numpy-1.1.tar.gz").st_size,
            ),
        ),
    )


@pytest.mark.asyncio
async def test_get_project_page__not_found(simple_dir: Path) -> None:
    repo = LocalRepository(
        index_path=simple_dir,
    )
    with pytest.raises(
        errors.PackageNotFoundError,
        match="seaborn",
    ):
        await repo.get_project_page("seaborn")



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from datetime import datetime
import os
import pathlib
import typing
from unittest import mock

import httpx
import pytest
import pytest_asyncio
from pytest_httpx import HTTPXMock

from ... import model
from ...components.http_cached import CachedHttpRepository


@pytest_asyncio.fixture
async def repository(
    tmp_path: pathlib.Path,
) -> typing.AsyncGenerator[CachedHttpRepository, None]:
    async with httpx.AsyncClient() as client:
        yield CachedHttpRepository(
            url="https://example.com/simple/",
            http_client=client,
            cache_path=tmp_path,
        )


@pytest.mark.asyncio
async def test_fetch_simple_page__cache_miss(
    repository: CachedHttpRepository,
    httpx_mock: HTTPXMock,
) -> None:
    httpx_mock.add_response(
        content="new-body",
        headers={
            "Content-Type": "new-type",
            "ETag": "new-etag",
        },
    )

    body, content_type = await repository._fetch_simple_page("http://url")
    assert body == "new-body"
    assert content_type == "new-type"

    cached = repository._get_from_cache("http://url")
    assert cached == "new-etag,new-type,new-body"


@pytest.mark.asyncio
async def test_fetch_simple_page__cache_hit_not_modified(
    repository: CachedHttpRepository,
    httpx_mock: HTTPXMock,
) -> None:
    httpx_mock.add_response(status_code=304)

    repository._save_to_cache("http://url", "stored-etag,stored-type,stored-body")
    body, content_type = await repository._fetch_simple_page("http://url")
    assert body == "stored-body"
    assert content_type == "stored-type"


@pytest.mark.asyncio
async def test_fetch_simple_page__cache_hit_modified(
    repository: CachedHttpRepository,
    httpx_mock: HTTPXMock,
) -> None:
    httpx_mock.add_response(
        content="new-body",
        headers={
            "Content-Type": "new-type",
            "ETag": "new-etag",
        },
    )
    repository._save_to_cache("http://url", "stored-etag,stored-type,stored-body")
    body, content_type = await repository._fetch_simple_page("http://url")
    assert body == "new-body"
    assert content_type == "new-type"
    assert repository._get_from_cache("http://url") == "new-etag,new-type,new-body"


@pytest.mark.asyncio
async def test_fetch_simple_page__cache_hit_source_unreachable(
    repository: CachedHttpRepository,
    httpx_mock: HTTPXMock,
) -> None:
    repository._save_to_cache("url", "stored-etag,stored-type,stored-body")
    httpx_mock.add_exception(httpx.RequestError("error"))
    body, content_type = await repository._fetch_simple_page("url")
    assert body == "stored-body"
    assert content_type == "stored-type"
    assert repository._get_from_cache("url") == "stored-etag,stored-type,stored-body"


@pytest.mark.asyncio
async def test_fetch_simple_page__cache_miss_source_unreachable(
    repository: CachedHttpRepository,
    httpx_mock: HTTPXMock,
) -> None:
    httpx_mock.add_exception(httpx.RequestError("error"))
    with pytest.raises(httpx.RequestError, match="error"):
        await repository._fetch_simple_page("url")


@pytest.mark.asyncio
async def test_get_project_page__cached(
    repository: CachedHttpRepository,
    httpx_mock: HTTPXMock,
) -> None:
    repository._save_to_cache(
        "https://example.com/simple/project/", "stored-etag,text/html," + """
        <a href="test1.whl#hash=test_hash">test1.whl</a>
        <a href="http://test2.whl">test2.whl</a>
    """,
    )
    httpx_mock.add_response(status_code=304)
    response = await repository.get_project_page("project")

    assert response == model.ProjectDetail(
        name="project",
        meta=model.Meta(
            api_version="1.0",
        ),
        files=(
            model.File(
                filename="test1.whl",
                url="https://example.com/simple/project/test1.whl",
                hashes={"hash": "test_hash"},
            ),
            model.File(
                filename="test2.whl",
                url="http://test2.whl",
                hashes={},
            ),
        ),
    )


@pytest.mark.asyncio
async def test_get_project_list__cached(
    repository: CachedHttpRepository,
    httpx_mock: HTTPXMock,
) -> None:
    repository._save_to_cache(
        "https://example.com/simple/", "stored-etag,text/html," + """
        <a href="/p1/">p1</a>
        <a href="/p2/">p2</a>
    """,
    )
    httpx_mock.add_response(status_code=304)

    resp = await repository.get_project_list()
    assert resp == model.ProjectList(
        meta=model.Meta(
            api_version="1.0",
        ),
        projects=frozenset([
            model.ProjectListElement(name="p1"),
            model.ProjectListElement(name="p2"),
        ]),
    )


def test_update_access_time(repository: CachedHttpRepository) -> None:
    repository._save_to_cache("url", "content")
    with mock.patch(
        "simple_repository.components.http_cached.datetime",
        mock.Mock(
            now=mock.Mock(return_value=datetime.fromisoformat("2006-07-09")),
            fromisoformat=datetime.fromisoformat,
            spec=datetime,
        ),
    ):
        repository._get_from_cache("url")

    assert os.path.getmtime(repository._cache_path / "url") == datetime.fromisoformat("2006-07-09").timestamp()



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import datetime
from logging import Logger, getLogger
import os
import pathlib
import shutil
import uuid

import httpx

from .. import errors, model, utils
from .core import RepositoryContainer, SimpleRepository


class ResourceCacheRepository(RepositoryContainer):
    """
    A cache for resources based on etags. It stores temporarily
    resources with an asigned etag on the local disk.
    When fallback_to_cache is enabled (default), then a cached resource
    will be returned if the source repository is unavailable. Used in
    the right context, this can be used for example to maintain a functional
    repository for previously seen responses when PyPI.org is down.
    """
    def __init__(
        self,
        source: SimpleRepository,
        cache_path: pathlib.Path,
        http_client: httpx.AsyncClient | None = None,
        logger: Logger = getLogger(__name__),
        fallback_to_cache: bool = True,
    ) -> None:
        super().__init__(source)
        self._cache_path = cache_path.resolve()
        self._tmp_path = self._cache_path / ".incomplete"
        self._tmp_path.mkdir(parents=True, exist_ok=True)
        self._http_client = http_client or httpx.AsyncClient()
        self._logger = logger
        self._fallback_to_cache = fallback_to_cache

    async def get_resource(
        self,
        project_name: str,
        resource_name: str,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.Resource:
        """
        Get a resource from the cache. If it is not present in the
        cache, or its etag has expired, retrieve it from the source repository.
        If it's a remote resource, download the resource and cache it.
        """
        project_dir = (self._cache_path / project_name).resolve()
        resource_path = (project_dir / resource_name).resolve()
        resource_info_path = resource_path.with_suffix(resource_path.suffix + ".info")

        # Ensures that the requested resource is contained
        # in the cache directory to avoid path traversal.
        if (
            not resource_path.is_relative_to(self._cache_path) or
            not project_dir.is_relative_to(self._cache_path)
        ):
            raise ValueError(f"{resource_path} is not contained in {self._cache_path}")

        project_dir.mkdir(exist_ok=True)

        # Require the resource upstream, if available use the cached etag.
        cache_etag = resource_info_path.read_text() if resource_info_path.is_file() else None
        if cache_etag:
            context = request_context.context | {"etag": cache_etag}
        else:
            context = request_context.context
        new_request_context = model.RequestContext(
            repository=request_context.repository,
            context=context,
        )
        try:
            resource = await super().get_resource(
                project_name,
                resource_name,
                request_context=new_request_context,
            )
        except model.NotModified:
            # The upstream repository serves the same content that has been cached.
            # If the request also provides the same etag, raise NotModified,
            # otherwhise return the locally cached resource.
            if cache_etag == request_context.context.get("etag"):
                raise
            return self._cached_resource(
                resource_path=resource_path,
                resource_info_path=resource_info_path,
                context=model.Context(etag=cache_etag) if cache_etag else model.Context(),
            )
        except errors.SourceRepositoryUnavailable:
            if not self._fallback_to_cache:
                raise
            # If the source repository behaves incorrectly and we
            # have a cached artifact, return it to the client.
            self._logger.error(
                f"Upstream unavailable, served cached {project_name}:{resource_name}",
            )
            if cache_etag:
                return self._cached_resource(
                    resource_path=resource_path,
                    resource_info_path=resource_info_path,
                    context=model.Context(etag=cache_etag),
                )
            else:
                raise

        upstream_etag = resource.context.get("etag")
        if not upstream_etag or resource.to_cache is False:
            # Only cache a resource if has an etag and to_cache is set to True.
            return resource

        if upstream_etag != cache_etag:
            await self._store_resource(
                resource=resource,
                upstream_etag=upstream_etag,
                resource_path=resource_path,
                resource_info_path=resource_info_path,
            )

        return self._cached_resource(
            resource_path=resource_path,
            resource_info_path=resource_info_path,
            context=resource.context,
        )

    async def _store_resource(
        self,
        resource: model.Resource,
        upstream_etag: str,
        resource_path: pathlib.Path,
        resource_info_path: pathlib.Path,
    ) -> None:
        if isinstance(resource, model.HttpResource):
            # The upstream resource changed or no cached version is available.
            # Fetch the resource and cache it and its etag.
            timestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
            dest_file = self._tmp_path / f"{timestamp}_{uuid.uuid4().hex}"
            await utils.download_file(
                download_url=resource.url,
                dest_file=dest_file,
                http_client=self._http_client,
            )
            dest_file.rename(resource_path)
        elif isinstance(resource, model.TextResource):
            resource_path.write_text(resource.text)
        elif isinstance(resource, model.LocalResource):
            shutil.copy(resource.path, resource_path)
        else:
            raise ValueError(f"Unknow resource type: {type(resource)}.")
        resource_info_path.write_text(upstream_etag)

    def _cached_resource(
        self,
        resource_path: pathlib.Path,
        resource_info_path: pathlib.Path,
        context: model.Context,
    ) -> model.LocalResource:
        self._update_last_access(resource_info_path)
        local_resource = model.LocalResource(path=resource_path)
        local_resource.context.update(context)
        return local_resource

    def _update_last_access(
        self,
        resource_info_path: pathlib.Path,
    ) -> None:
        """
        Store the last access as the access and modified times of the file.
        That information will be used to delete unused files in the cache.
        """
        now = datetime.datetime.now().timestamp()

        os.utime(resource_info_path, (now, now))



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from dataclasses import replace
from datetime import timedelta
from urllib.parse import urljoin

import httpx

from .. import errors, model, parser, utils
from .core import SimpleRepository


class HttpRepository(SimpleRepository):
    """Proxy of a remote simple repository"""

    def __init__(
        self,
        url: str,
        http_client: httpx.AsyncClient | None = None,
        connection_timeout: timedelta = timedelta(seconds=15),
    ):
        self.source_url = url
        self._http_client = http_client or httpx.AsyncClient()
        self.downstream_content_types = ", ".join([
            "application/vnd.pypi.simple.v1+json",
            "application/vnd.pypi.simple.v1+html;q=0.2",
            "text/html;q=0.01",
        ])
        self._connection_timeout = connection_timeout

    async def _fetch_simple_page(
        self,
        page_url: str,
    ) -> tuple[str, str]:
        """Retrieves a simple page from the given url.
        Returns the body and the content type received.
        """
        headers = {"Accept": self.downstream_content_types}
        response = await self._http_client.get(
            url=page_url,
            headers=headers,
            timeout=self._connection_timeout.total_seconds(),
        )
        response.raise_for_status()
        body = response.text
        content_type: str = response.headers.get("content-type", "")
        return body, content_type

    async def get_project_page(
        self,
        project_name: str,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.ProjectDetail:
        page_url = urljoin(self.source_url, f"{project_name}/")
        try:
            body, content_type = await self._fetch_simple_page(page_url)
        except httpx.HTTPError as e:
            # If the status_code is 404, the source repository is working correctly, but
            # the requested resource is not available. Any other 4xx or 5xx error code is
            # treated as a source repository misbehaviour
            if isinstance(e, httpx.HTTPStatusError) and e.response.status_code == 404:
                raise errors.PackageNotFoundError(
                    package_name=project_name,
                ) from e
            # This code path also includes connection failures or timeouts.
            raise errors.SourceRepositoryUnavailable() from e

        if (
            "application/vnd.pypi.simple.v1+html" in content_type or
            "text/html" in content_type or not content_type
        ):
            project_page = parser.parse_html_project_page(body, project_name)
        elif "application/vnd.pypi.simple.v1+json" in content_type:
            project_page = parser.parse_json_project_page(body)
        else:
            raise errors.UnsupportedSerialization(content_type)

        # Make the URLs in the project page absolute, such that they can be
        # resolved upstream without knowing the original source URLs.
        files = tuple(
            replace(file, url=utils.url_absolutizer(file.url, page_url))
            for file in project_page.files
        )
        project_page = replace(project_page, files=files)
        return project_page

    async def get_project_list(
        self,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.ProjectList:
        try:
            body, content_type = await self._fetch_simple_page(self.source_url)
        except httpx.HTTPError as e:
            raise errors.SourceRepositoryUnavailable() from e

        if (
            "application/vnd.pypi.simple.v1+html" in content_type or
            "text/html" in content_type or not content_type
        ):
            return parser.parse_html_project_list(body)
        elif "application/vnd.pypi.simple.v1+json" in content_type:
            return parser.parse_json_project_list(body)

        raise errors.UnsupportedSerialization(content_type)

    async def get_resource(
        self,
        project_name: str,
        resource_name: str,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.Resource:
        try:
            project_page = await self.get_project_page(
                project_name,
                request_context=request_context,
            )
        except errors.PackageNotFoundError:
            raise errors.ResourceUnavailable(resource_name)

        resource: model.HttpResource | None = None
        if resource_name.endswith(".metadata"):
            resource = await self.get_metadata(project_page, resource_name)
        else:
            for file in project_page.files:
                if resource_name == file.filename:
                    resource = model.HttpResource(url=file.url)
                    break

        if not resource:
            raise errors.ResourceUnavailable(resource_name)

        try:
            resp = await self._http_client.head(
                url=resource.url,
                timeout=self._connection_timeout.total_seconds(),
            )
            resp.raise_for_status()
        except httpx.HTTPError as e:
            raise errors.SourceRepositoryUnavailable() from e
        if etag := resp.headers.get("ETag"):
            if etag == request_context.context.get("etag"):
                # If the etag served from the source repository
                # matches the one in the request raise NotModified
                raise model.NotModified()
            resource.context["etag"] = etag

        return resource

    async def get_metadata(
        self,
        project_page: model.ProjectDetail,
        resource_name: str,
    ) -> model.HttpResource:
        distribution_name = resource_name.removesuffix(".metadata")
        for file in project_page.files:
            if distribution_name == file.filename and file.dist_info_metadata:
                return model.HttpResource(url=file.url + ".metadata")
        raise errors.ResourceUnavailable(resource_name)



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from html import escape
import json
import typing
from typing import Union

import packaging.utils
import packaging.version

from .content_negotiation import Format
from .model import File, ProjectDetail, ProjectList


class Serializer(typing.Protocol):
    def serialize_project_page(self, page: ProjectDetail) -> str:
        ...

    def serialize_project_list(self, page: ProjectList) -> str:
        ...


class SerializerJsonV1(Serializer):
    def serialize_project_page(self, page: ProjectDetail) -> str:
        project_page_dict = {
            "meta": {
                "api-version": page.meta.api_version,
            },
            "name": page.name,
            "files": [
                self.standardize_file(
                    file=file,
                    version=packaging.version.Version(page.meta.api_version),
                )
                for file in page.files
            ],
        }
        if page.versions is not None:
            project_page_dict["versions"] = list(page.versions)
        return json.dumps(project_page_dict)

    def serialize_project_list(self, page: ProjectList) -> str:
        list_dict = {
           "meta": {"api-version": page.meta.api_version},
           "projects": [
                {"name": elem.name} for elem in page.projects
           ],
        }
        return json.dumps(list_dict)

    def standardize_file(
        self,
        file: File,
        version: packaging.version.Version,
    ) -> dict[str, typing.Any]:
        file_dict: dict[str, typing.Any] = {
            "filename": file.filename,
            "url": file.url,
            "hashes": file.hashes,
        }
        if file.requires_python is not None:
            file_dict["requires-python"] = file.requires_python
        # From PEP-714: The PEP 658 metadata, when used in the PEP 691
        # JSON representation of the Simple API, MUST be emitted using
        # the key core-metadata, with the supported values remaining the same.
        if file.dist_info_metadata is not None:
            file_dict["core-metadata"] = file.dist_info_metadata
        if file.gpg_sig is not None:
            file_dict["gpg-sig"] = file.gpg_sig
        if file.yanked is not None:
            file_dict["yanked"] = file.yanked
        if version >= packaging.version.Version("1.1"):
            file_dict["size"] = file.size
            if file.upload_time is not None:
                file_dict["upload-time"] = file.upload_time.strftime("%Y-%m-%dT%H:%M:%SZ")
        return file_dict


class SerializerHtmlV1(Serializer):
    SIMPLE_PROJECT_HEADER = """<!DOCTYPE html>
    <html>
    <head>
        <meta name="pypi:repository-version" content="{api_version}">
        <title>Links for {project_name}</title>
    </head>
    <body>
    <h1>Links for {project_name}</h1>
"""
    SIMPLE_PROJECT_LINK = '<a {attributes}>{file_name}</a><br/>\n'
    SIMPLE_PROJECT_FOOTER = "</body>\n</html>"
    SIMPLE_INDEX_HEADER = """<!DOCTYPE html>
    <html>
    <head>
        <meta name="pypi:repository-version" content="{api_version}">
        <title>Simple index</title>
    </head>
    <body>
"""
    SIMPLE_INDEX_PROJECT_LINK = '<a href="{href}">{project}</a><br/>\n'
    SIMPLE_INDEX_FOOTER = "</body>\n</html>"

    def serialize_project_page(self, page: ProjectDetail) -> str:
        project_page_html = [
            self.SIMPLE_PROJECT_HEADER.format(
                api_version=page.meta.api_version,
                project_name=page.name,
            ),
        ]

        for file in page.files:
            project_page_html.append(self._serialize_file(file))

        project_page_html.append(self.SIMPLE_PROJECT_FOOTER)
        return "".join(project_page_html)

    def serialize_project_list(self, page: ProjectList) -> str:
        project_list_html = [
            self.SIMPLE_INDEX_HEADER.format(
                api_version=page.meta.api_version,
            ),
        ]
        for project_name in page.projects:
            project_list_html.append(
                self.SIMPLE_INDEX_PROJECT_LINK.format(
                    project=project_name.name,
                    # Using relative paths for project page urls.
                    href=packaging.utils.canonicalize_name(project_name.name) + "/",
                ),
            )
        project_list_html.append(self.SIMPLE_INDEX_FOOTER)
        return "".join(project_list_html)

    def _serialize_file(self, file: File) -> str:
        url = file.url
        attributes = []
        if file.hashes:
            hash_fun = "sha256" if "sha256" in file.hashes else next(iter(file.hashes))
            hash_value = file.hashes[hash_fun]
            fragment = f"{hash_fun}={hash_value}"
            url = f"{url}#{fragment}"

        attributes.append(f'href="{url}"')

        if file.requires_python:
            # From PEP 503: In the attribute value, < and > have to be HTML
            # encoded as &lt; and &gt;, respectively.
            attributes.append(f'data-requires-python="{escape(file.requires_python)}"')

        # From PEP 658: The repository SHOULD provide the hash of the Core Metadata file as the
        # data-dist-info-metadata attribute’s value using the syntax <hashname>=<hashvalue>,
        # where <hashname> is the lower cased name of the hash function used, and <hashvalue>
        # is the hex encoded digest. The repository MAY use true as the attribute’s value
        # if a hash is unavailable.

        # From PEP 714: The PEP 658 metadata, when used in the HTML representation of the Simple
        # API, MUST be emitted using the attribute name data-core-metadata, with the supported
        # values remaining the same.
        if file.dist_info_metadata:
            if file.dist_info_metadata is True:
                attributes.append('data-core-metadata="true"')
            else:
                hash_fun = (
                    "sha256" if "sha256" in file.dist_info_metadata
                    else next(iter(file.dist_info_metadata))
                )
                hash_value = file.dist_info_metadata[hash_fun]
                attributes.append(f'data-core-metadata="{hash_fun}={hash_value}"')

        # From PEP 592: The value of the data-yanked attribute, if present, is an arbitrary
        # string that represents the reason for why the file has been yanked.
        # According to PEP 691, if the reason is not specified, the value of the yanked key
        # is set to True and never to an empty string.
        if file.yanked:
            if file.yanked is True:
                attributes.append('data-yanked=""')
            else:
                attributes.append(f'data-yanked="{file.yanked}"')

        # From PEP 503: A repository MAY include a data-gpg-sig attribute on a file link with
        # a value of either true or false to indicate whether or not there is a GPG signature.
        if file.gpg_sig:
            attributes.append('data-gpg-sig="true"')
        elif file.gpg_sig is False:
            attributes.append('data-gpg-sig="false"')

        attributes_string = " ".join(attributes)

        return self.SIMPLE_PROJECT_LINK.format(
            file_name=file.filename,
            attributes=attributes_string,
        )


serializers: dict[Format, Serializer] = {
    Format.JSON_V1: SerializerJsonV1(),
    Format.HTML_V1: SerializerHtmlV1(),
    Format.HTML_LEGACY: SerializerHtmlV1(),
}


def serialize(page: Union[ProjectDetail, ProjectList], format: Format) -> str:
    serializer = serializers.get(format)
    if serializer is None:
        raise ValueError("Unsupported format")

    if isinstance(page, ProjectList):
        return serializer.serialize_project_list(page)
    elif isinstance(page, ProjectDetail):
        return serializer.serialize_project_page(page)
    else:
        raise ValueError("Unsupported page type")



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from enum import Enum

from . import errors


class Format(Enum):
    JSON_V1: str = "application/vnd.pypi.simple.v1+json"
    HTML_V1: str = "application/vnd.pypi.simple.v1+html"
    HTML_LEGACY: str = "text/html"


def select_response_format(content_type: str) -> Format:
    if not content_type:
        return Format.HTML_LEGACY

    content_type_tokens = content_type.split(",")
    requested_formats = []

    for token in content_type_tokens:
        token_pieces = token.replace(" ", "").split(";")
        q = 1.0
        if len(token_pieces) == 2:
            q = float(token_pieces[1].removeprefix("q="))
        requested_formats.append(
            (token_pieces[0], q),
        )
    # Requested content-types sorted by q value
    sorted_formats = sorted(
        requested_formats, key=lambda f: f[1], reverse=True,
    )
    for form, _ in sorted_formats:
        if form == "*/*":
            return Format.HTML_LEGACY
        try:
            return Format(form)
        except ValueError:
            pass

    raise errors.UnsupportedSerialization(content_type)



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import enum
import posixpath

import packaging.utils
from packaging.version import InvalidVersion, Version


def split_sdist_filename(path: str) -> tuple[str, str]:
    """
    Like os.path.splitext, but take off .tar too.
    Standard functions like splitext or pathlib suffixes
    will fail to extract the extention of sdists
    like numpy-1.0.0.tar.gz
    """
    base, ext = posixpath.splitext(path)
    if base.lower().endswith(".tar"):
        ext = base[-4:] + ext
        base = base[:-4]
    return base, ext


class PackageFormat(enum.Enum):
    WHEEL: str = "wheel"
    SDIST: str = "sdist"
    OTHER: str = "other format"


def extract_package_format(filename: str) -> PackageFormat:
    _, file_format = split_sdist_filename(filename)
    if file_format == ".whl":
        return PackageFormat.WHEEL
    if file_format in ('.zip', '.tar.gz', '.tar.bz2', '.tar.xz', '.tar.Z', '.tar'):
        return PackageFormat.SDIST
    return PackageFormat.OTHER


def extract_version_from_fragment(fragment: str, project_name: str) -> str:
    def find_version_start() -> int:
        pieces = fragment.split("-")
        for i in range(len(pieces)):
            candidate = "-".join(pieces[0:i])
            if packaging.utils.canonicalize_name(candidate) == project_name:
                return len(candidate)
        raise ValueError(f"{fragment} does not match {project_name}")
    return fragment[find_version_start() + 1:]


def extract_package_version(filename: str, project_name: str) -> str:
    if extract_package_format(filename) == PackageFormat.WHEEL:
        return filename.split('-')[1]
    else:
        fragment, _ = split_sdist_filename(filename)
        return extract_version_from_fragment(fragment, project_name)


def safe_version(filename: str, project_name: str) -> Version:
    try:
        return Version(
            version=extract_package_version(
                filename=filename,
                project_name=project_name,
            ),
        )
    except (ValueError, InvalidVersion):
        return Version('0.0rc0')




# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import pytest

from ... import errors, model
from ...components.core import RepositoryContainer
from .fake_repository import FakeRepository


@pytest.mark.asyncio
async def test_get_project_page() -> None:
    repository = RepositoryContainer(
        FakeRepository(
            project_pages=[
                model.ProjectDetail(model.Meta("1.0"), "numpy", files=()),
            ],
        ),
    )
    result = await repository.get_project_page("numpy")
    assert result == model.ProjectDetail(model.Meta("1.0"), "numpy", files=())
    with pytest.raises(errors.PackageNotFoundError):
        await repository.get_project_page("pandas")


@pytest.mark.asyncio
async def test_get_project_list() -> None:
    repository = RepositoryContainer(
        FakeRepository(
            project_list=model.ProjectList(
                meta=model.Meta("1.0"),
                projects=frozenset([model.ProjectListElement("numpy")]),
            ),
        ),
    )
    result = await repository.get_project_list()
    assert result.projects == frozenset([model.ProjectListElement("numpy")])


@pytest.mark.asyncio
async def test_get_resource() -> None:
    repository = RepositoryContainer(
        FakeRepository(
            resources={
                "numpy.whl": model.HttpResource("numpy_url"),
            },
        ),
    )
    result = await repository.get_resource("numpy", "numpy.whl")
    assert isinstance(result, model.HttpResource)
    assert result.url == "numpy_url"



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

"""
Models to represent a Simple package index

Essential reading (in order):

 * https://peps.python.org/pep-0503/  (html simple index)
 * https://peps.python.org/pep-0629/  (simple index versioning)
 * https://peps.python.org/pep-0691/  (json simple index)
 * https://peps.python.org/pep-0592/  (yank)
 * https://peps.python.org/pep-0658/  (metadata files)

In general, the model follows the latest PEPs wherever possible, and it is the
responsibility of the parsers and serializers to adapt to the newest standards.

In the comments of this model, we collate the relevant parts of the PEPs.
Direct PEP quotes look like:

    PEP-XXXX: Some directly copy & pasted text from the PEP, which somebody
              could find easily in the PEP with ctrl+F in the browser.

Where additional context is added to a quote, it will be inline within square
brackets, for example ``[additional context]``.

"""
from dataclasses import dataclass, field
from datetime import datetime
import pathlib
from typing import TYPE_CHECKING, Optional, TypedDict, Union

import packaging.utils
import packaging.version

from .packaging import safe_version

if TYPE_CHECKING:
    from . import SimpleRepository


@dataclass(frozen=True)
class File:
    """
    Simple representation of a distribution file.
    Defined in PEP-691: https://peps.python.org/pep-0691/
    """

    filename: str
    url: str

    # PEP-592: A dictionary mapping a hash name to a hex encoded digest of the file.
    # PEP-592: Limited to a len() of 1 in HTML
    # Additional hashes are not included in the HTML serialization.
    hashes: dict[str, str]
    requires_python: Optional[str] = None

    # PEP-691: An optional key that indicates that metadata for this file is available
    # PEP-691: Where this is present, it MUST be either a boolean to indicate
    #          if the file has an associated metadata file, or a dictionary mapping hash
    #          names to a hex encoded digest of the metadata’s hash.
    # PEP-658: The presence [in HTML] of the attribute indicates the distribution
    #          represented by the anchor tag MUST contain a Core Metadata file that
    #          will not be modified when the distribution is processed and/or installed
    # PEP-691: Limited to a len() of 1 in HTML
    # If the key is not "present", it will be None.
    # A maximum of one hash will be included the HTML serialization.
    dist_info_metadata: Optional[Union[bool, dict[str, str]]] = None  # PEP-658

    # PEP-691: An optional key that acts a boolean to indicate if the file has an
    #          associated GPG signature or not.
    # PEP-691: If this key does not exist, then the signature may or may not exist.
    # A None value indicates the key does not exist (i.e. there may or may not be a sig file).
    gpg_sig: Optional[bool] = None

    # PEP-691: either a boolean to indicate if the file has been yanked, or a non empty,
    #          but otherwise arbitrary, string to indicate that a file
    #          has been yanked with a specific reason. If the yanked key is present and
    #          is a [JSON] truthy value, then it SHOULD be interpreted as indicating that
    #          the file pointed to by the url field has been “Yanked”
    # PEP-592: The value of the data-yanked attribute [in HTML], if present, is an
    #          arbitrary string [including falsy ones such as "false"] that represents
    #          the reason for why the file has been yanked.
    # Note that the string "false" is a valid yank reason in both JSON and HTML.
    yanked: Optional[Union[bool, str]] = None

    # PEP-700: This field is mandatory. It MUST contain an integer which is the file size in bytes.
    size: Optional[int] = None

    # PEP-700: This field is optional. If present, it MUST contain a valid ISO 8601 date/time
    #          string, in the format yyyy-mm-ddThh:mm:ss.ffffffZ, which represents the time the
    #          file was uploaded to the index. As indicated by the Z suffix, the upload time
    #          MUST use the UTC timezone.
    upload_time: Optional[datetime] = None

    def __post_init__(self) -> None:
        if self.yanked == "":
            raise ValueError("The yanked attribute may not be an empty string")


@dataclass(frozen=True)
class Meta:
    """Responses metadata defined in PEP-629:
    https://peps.python.org/pep-0629/
    """
    api_version: str


@dataclass(frozen=True)
class ProjectDetail:
    """Model of a project page as described in PEP-691"""
    meta: Meta
    name: str
    files: tuple[File, ...]
    # PEP-700: An additional key, versions MUST be present at the top level, in addition to the
    #          keys name, files and meta defined in PEP 691. This key MUST contain a list of version
    #          strings specifying all of the project versions uploaded for this project.
    #
    # This field is automatically calculated when a ProjectDetail is created with api_version>=1.1.
    versions: Optional[set[str]] = field(init=False)

    def __post_init__(self) -> None:
        api_version = packaging.version.Version(self.meta.api_version)
        if api_version >= packaging.version.Version("1.1"):
            for file in self.files:
                if file.size is None:
                    raise ValueError(
                        "SimpleAPI>=1.1 requires the size field to be set for all the files.",
                    )
            versions = {
                str(safe_version(file.filename, self._normalized_name))
                for file in self.files
            }
        else:
            versions = None
        object.__setattr__(self, "versions", versions)

    @property
    def _normalized_name(self) -> str:
        return packaging.utils.canonicalize_name(self.name)


@dataclass(frozen=True)
class ProjectListElement:
    name: str  # not necessarily normalized.

    @property
    def normalized_name(self) -> str:
        return packaging.utils.canonicalize_name(self.name)


@dataclass(frozen=True)
class ProjectList:
    """Model of the project list as described in PEP-691"""
    meta: Meta
    projects: frozenset[ProjectListElement]


class Context(TypedDict, total=False):
    etag: str


@dataclass(frozen=True)
class Resource:
    context: Context = field(default_factory=lambda: Context(), kw_only=True)
    # If this attribute is set to False, cache components will ignore this resource
    to_cache: bool = field(default=True, kw_only=True)


@dataclass(frozen=True)
class RequestContext:
    repository: "SimpleRepository"
    # TODO: Worry that context is mutable.
    context: dict[str, str] = field(default_factory=dict)

    # Provider a default context which can be used in all signatures using RequestContext.
    # By default, if not specified, the default request context will be the one containing the
    # repository of the originating call (i.e. the repository upon which you call a method is the
    # one that is injected into the context, and passed down to each subsequent (nested) request.
    # We know that None isn't a RequestContext instance... as a result of this, every user
    # of RequestContext should handle this. RepositorySource automatically transforms this
    # to a sensible incoming request (see RepositorySource._build_request_context).
    DEFAULT: "RequestContext" = None  # type: ignore[assignment]


@dataclass(frozen=True)
class LocalResource(Resource):
    path: pathlib.Path


@dataclass(frozen=True)
class HttpResource(Resource):
    url: str


@dataclass(frozen=True)
class TextResource(Resource):
    text: str


class NotModified(Exception):
    """
    Exception to be used if the requested resource
    is not changed with respect to the request ETag.
    """



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

"""Test the custom html parser that's used for parsing (remote) Simple index html pages

Currently, there are no tests written for parsing invalid HTML.
This is because it (html.parser.HTMLParser) does not throw errors, but gives strange results instead
"""

from ..html_parser import HTMLElement, SimpleHTMLParser


def test_parser() -> None:
    parser = SimpleHTMLParser()
    data = """<!DOCTYPE html>
<html>
  <head>
    <meta name="pypi:repository-version" content="1.0"/>
    <title>Title</title>
  </head>
  <body>
    <h1>Header</h1>
    <a href="../../hammer/hammer-1.0.tar.gz">hammer-1.0.tar.gz</a><br/>
    <a href="../../shovel/shovel-1.0.whl" extra-attr>shovel-1.0.whl</a><br/>
    <a href="../../drill/drill-1.0.whl" extra-attr="true">drill-1.0.whl</a><br/>
  </body>
</html>
"""
    parser.feed(data)
    assert parser.declaration == "DOCTYPE html"
    assert parser.elements == [
        HTMLElement("html", {}),
        HTMLElement("head", {}),
        HTMLElement("meta", {"name": "pypi:repository-version", "content": "1.0"}),
        HTMLElement("title", {}, "Title"),
        HTMLElement("body", {}),
        HTMLElement("h1", {}, "Header"),
        HTMLElement("a", {"href": "../../hammer/hammer-1.0.tar.gz"}, "hammer-1.0.tar.gz"),
        HTMLElement("br", {}),
        emtpy_attr := HTMLElement("a", {"href": "../../shovel/shovel-1.0.whl", "extra-attr": None}, "shovel-1.0.whl"),
        HTMLElement("br", {}),
        HTMLElement("a", {"href": "../../drill/drill-1.0.whl", "extra-attr": "true"}, "drill-1.0.whl"),
        HTMLElement("br", {}),
    ]
    assert str(emtpy_attr) == '<a href="../../shovel/shovel-1.0.whl" extra-attr="">shovel-1.0.whl</a>'



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import json
import pathlib
import typing
from urllib.parse import urljoin, urlparse

import httpx

from . import errors


def url_absolutizer(url: str, url_base: str) -> str:
    """Converts a relative url into an absolute one"""
    if not urlparse(url).scheme:
        return urljoin(url_base, url)
    return url


def load_config_json(json_file: pathlib.Path) -> dict[typing.Any, typing.Any]:
    try:
        json_config = json.loads(json_file.read_text())
    except json.JSONDecodeError as e:
        raise errors.InvalidConfigurationError("Invalid json file") from e
    except FileNotFoundError as e:
        raise errors.InvalidConfigurationError("Configuration file not found") from e
    if not isinstance(json_config, dict):
        raise errors.InvalidConfigurationError(
            f"Invalid configuration file. {str(json_file)}"
            " must contain a dictionary.",
        )
    return json_config


async def download_file(
    download_url: str,
    dest_file: pathlib.Path,
    http_client: httpx.AsyncClient,
    chunk_size: int = 1024 * 64,
) -> None:
    with dest_file.open('wb') as file:
        async with http_client.stream("GET", download_url) as data:
            async for chunk in data.aiter_bytes(chunk_size):
                file.write(chunk)



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from dataclasses import replace
import pathlib
import re
import tempfile
import zipfile

import httpx
from packaging.utils import canonicalize_name

from .. import errors, model, utils
from .core import RepositoryContainer, SimpleRepository

metadata_regex = re.compile(r'^(.*)-.*\.dist-info/METADATA$')


class MetadataInjectorRepository(RepositoryContainer):
    """Adds PEP-658 support to a simple repository. If not already specified,
    sets the dist-info metadata for all wheels packages in a project page.
    Metadata is extracted from the wheels on the fly and cached for later use.
    """
    def __init__(
        self,
        source: SimpleRepository,
        http_client: httpx.AsyncClient | None = None,
    ) -> None:
        self._http_client = http_client or httpx.AsyncClient()
        super().__init__(source)

    async def get_project_page(
        self,
        project_name: str,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.ProjectDetail:
        return self._add_metadata_attribute(
            await super().get_project_page(project_name, request_context=request_context),
        )

    async def get_resource(
        self,
        project_name: str,
        resource_name: str,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.Resource:
        try:
            # Attempt to get the resource from upstream.
            return await super().get_resource(
                project_name,
                resource_name,
                request_context=request_context,
            )
        except errors.ResourceUnavailable:
            if not resource_name.endswith(".metadata"):
                # If we tried to get a resource that wasn't a .metadata one, and it failed,
                # propagate the error.
                raise

        # The resource doesn't exist upstream, and looks like a metadata file has been
        # requested. Let's try to fetch the underlying resource and compute the metadata.

        # Get hold of the actual artefact from which we want to extract
        # the metadata.
        resource = await request_context.repository.get_resource(
            project_name, resource_name.removesuffix(".metadata"),
            request_context=request_context,
        )
        if isinstance(resource, model.HttpResource):
            try:
                metadata = await self._download_metadata(
                    package_name=resource_name.removesuffix(".metadata"),
                    download_url=resource.url,
                    http_client=self._http_client,
                )
            except ValueError as e:
                # If we can't get hold of the metadata from the file then raise
                # a resource unavailable.
                raise errors.ResourceUnavailable(resource_name) from e
        elif isinstance(resource, model.LocalResource):
            try:
                metadata = self._get_metadata_from_package(resource.path)
            except ValueError as e:
                raise errors.ResourceUnavailable(resource_name) from e
        else:
            raise errors.ResourceUnavailable(
                resource_name.removesuffix(".metadata"),
                "Unable to fetch the resource needed to extract the metadata.",
            )

        metadata_resource = model.TextResource(
            text=metadata,
        )
        if etag := resource.context.get("etag"):
            # Use the same etag as the one that identifies the package.
            # In this way, if that package changes, also the metadata will be invalidated.
            metadata_resource.context["etag"] = etag
        return metadata_resource

    def _get_metadata_from_wheel(self, package_path: pathlib.Path) -> str:
        package_tokens = package_path.name.split('-')
        if len(package_tokens) < 2:
            raise ValueError(
                f"Filename {package_path.name} is not normalized according to PEP-427",
            )
        distribution = canonicalize_name(package_tokens[0])
        # Package consumer, when extracting metadata, should tolerate small differences
        # respecting what is strictly described in PEP-427, for reference see:
        # https://packaging.python.org/en/latest/specifications/binary-distribution-format/
        try:
            with zipfile.ZipFile(package_path, 'r') as ziparchive:
                for file in ziparchive.namelist():
                    if not (match := metadata_regex.match(file)):
                        continue
                    if canonicalize_name(match.group(1)) == distribution:
                        return ziparchive.read(file).decode()
                raise errors.InvalidPackageError(
                    "Provided wheel doesn't contain a metadata file.",
                )
        except (zipfile.BadZipFile, zipfile.LargeZipFile) as e:
            raise errors.InvalidPackageError(
                "Unable to decompress the provided wheel.",
            ) from e

    def _get_metadata_from_package(self, package_path: pathlib.Path) -> str:
        if package_path.name.endswith('.whl'):
            return self._get_metadata_from_wheel(package_path)
        raise ValueError("Package provided is not a wheel")

    async def _download_metadata(
        self,
        package_name: str,
        download_url: str,
        http_client: httpx.AsyncClient,
    ) -> str:
        with tempfile.TemporaryDirectory() as tmpdir:
            pkg_path = pathlib.Path(tmpdir) / package_name
            await utils.download_file(download_url, pkg_path, http_client)
            return self._get_metadata_from_package(pkg_path)

    def _add_metadata_attribute(
        self,
        project_page: model.ProjectDetail,
    ) -> model.ProjectDetail:
        """Add the data-core-metadata to all the packages distributed as wheels"""
        files = []
        for file in project_page.files:
            if file.url and file.filename.endswith(".whl") and not file.dist_info_metadata:
                file = replace(file, dist_info_metadata=True)
            files.append(file)
        project_page = replace(project_page, files=tuple(files))
        return project_page



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from datetime import datetime
from html import unescape
import json
from typing import Optional, Union
from urllib.parse import quote, urldefrag

from . import html_parser
from .model import File, Meta, ProjectDetail, ProjectList, ProjectListElement


def parse_json_project_list(page: str) -> ProjectList:
    project_dict = json.loads(page)
    projects = frozenset(
        ProjectListElement(
            name=project.get("name"),
        ) for project in project_dict["projects"]
    )
    return ProjectList(
        meta=Meta(
            api_version=project_dict["meta"]["api-version"],
        ),
        projects=projects,
    )


def parse_html_project_list(page: str) -> ProjectList:
    parser = html_parser.SimpleHTMLParser()
    if not page.lower().lstrip().startswith("<!DOCTYPE html>"):
        # Temporary fix: https://github.com/pypa/pip/issues/10825
        page = "<!DOCTYPE html>\n" + page
    parser.feed(page)

    a_tags = (
        element for element in parser.elements
        if element.tag == "a"
    )

    projects = frozenset(
        ProjectListElement(
            name=element.content,
        )
        for element in a_tags
        if element.content is not None
    )

    return ProjectList(
        meta=Meta(
            api_version="1.0",
        ),
        projects=projects,
    )


def parse_json_project_page(body: str) -> ProjectDetail:
    page_dict = json.loads(body)

    files = []
    for file in page_dict["files"]:
        if date_string := file.get("upload-time"):
            # PEP-700: upload-time MUST contain a valid ISO 8601 date/time string, in the format
            # yyyy-mm-ddThh:mm:ss.ffffffZ, which represents the time the file was uploaded to
            # the index. The fractional seconds part of the timestamp is optional.
            try:
                upload_time = datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%SZ")
            except ValueError:
                upload_time = datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%S.%fZ")
        else:
            upload_time = None
        files.append(
            File(
                filename=file["filename"],
                # Temporary fix: Escape the URLs coming from the source
                # since Nexus is not escaping them correctly. The "safe" parameter
                # allows the escape function to be applied to the entire URL
                # and avoids escaping the special characters ":" and "/".
                url=quote(file["url"], safe=":/"),
                hashes=file["hashes"],
                requires_python=file.get("requires-python"),
                # PEP-714: Clients consuming the JSON represenation of the Simple API MUST
                #          read the PEP 658 metadata from the key core-metadata if it is present.
                dist_info_metadata=file.get("core-metadata"),
                gpg_sig=file.get("gpg-sig"),
                yanked=file.get("yanked"),
                size=file.get("size"),
                upload_time=upload_time,
            ),
        )

    return ProjectDetail(
        name=page_dict["name"],
        meta=Meta(
            api_version=page_dict["meta"]["api-version"],
        ),
        files=tuple(files),
    )


def parse_html_project_page(page: str, project_name: str) -> ProjectDetail:
    parser = html_parser.SimpleHTMLParser()
    if not page.lower().lstrip().startswith("<!DOCTYPE html>"):
        # Temporary fix: https://github.com/pypa/pip/issues/10825
        page = "<!DOCTYPE html>\n" + page
    parser.feed(page)

    files = []
    a_tags = (
        e for e in parser.elements if e.tag == "a"
    )

    for a_tag in a_tags:
        if (a_tag.content is None) or (a_tag.attrs.get("href") is None):
            continue

        hashes = {}
        url, fragment = urldefrag(a_tag.attrs["href"])

        if fragment:
            # PEP-503: The URL SHOULD include a hash in the form of a URL fragment with
            #          the following syntax: #<hashname>=<hashvalue>
            if '=' in fragment:
                hash_name, hash_value = str(fragment).split('=', 1)
                hashes[hash_name] = hash_value

        yanked: Optional[Union[bool, str]] = None
        if "data-yanked" in a_tag.attrs:
            if reason := a_tag.attrs.get("data-yanked"):
                # Note that the reason can equally be the string "false" and it is
                # still considered yanked.
                yanked = reason
            else:
                # The data-yanked value is not set or is an empty string, replace it with True.
                yanked = True

        dist_info_metadata: Optional[Union[bool, dict[str, str]]] = None
        # PEP-714: Clients consuming any of the HTML representations of the Simple API MUST
        #          read the PEP 658 metadata from the key data-core-metadata if it is present.
        if "data-core-metadata" in a_tag.attrs:
            # PEP-658: The repository SHOULD provide the hash of the Core Metadata file
            #          as the data-dist-info-metadata attribute’s value using
            #          the syntax <hashname>=<hashvalue>
            metadata_val = a_tag.attrs.get("data-core-metadata")
            if metadata_val is None:
                # data-core-metadata is set but doesn't have a value.
                dist_info_metadata = True
            else:
                metadata_attr_tokens = metadata_val.split("=", 1)
                if len(metadata_attr_tokens) == 2:
                    # the value of data-core-metadata can be parsed as <hash_fun>=<hash_val>.
                    dist_info_metadata = {metadata_attr_tokens[0]: metadata_attr_tokens[1]}
                else:
                    # the value of data-core-metadata is a placeholder. It doesn't follow
                    # the SHOULD recommendation, but it is still indicating that the
                    # metadata exists.
                    dist_info_metadata = True

        gpg_sig: Optional[bool] = None
        if gpg_sig_value := a_tag.attrs.get("data-gpg-sig"):
            # PEP-503: A repository MAY include a data-gpg-sig attribute on a file link with
            #          a value of either true or false to indicate whether or not there is a
            #          GPG signature. Repositories that do this SHOULD include it on every link.
            if gpg_sig_value == "true":
                gpg_sig = True
            elif gpg_sig_value == "false":
                gpg_sig = False

        requires_python: Optional[str] = None
        requires_python_attr = a_tag.attrs.get("data-requires-python")
        if requires_python_attr is not None:
            # PEP-503: A repository MAY include a data-requires-python attribute on a file link.
            #          This exposes the Requires-Python metadata field, specified in PEP 345, for
            #          the corresponding release. Where this is present, installer tools SHOULD
            #          ignore the download when installing to a Python version that doesn’t
            #          satisfy the requirement. In the attribute value, < and > have to be HTML
            #          encoded as &lt; and &gt;, respectively.
            requires_python = unescape(requires_python_attr)

        file = File(
            filename=a_tag.content,
            # Temporary fix: Escape the URLs coming from the source
            # since Nexus is not escaping them correctly. The "safe" parameter
            # allows the escape function to be applied to the entire URL
            # and avoids escaping the special characters ":" and "/".
            url=quote(str(url), safe=":/"),
            hashes=hashes,
            requires_python=requires_python,
            dist_info_metadata=dist_info_metadata,
            yanked=yanked,
            gpg_sig=gpg_sig,
        )

        files.append(file)

    return ProjectDetail(
        name=project_name,
        meta=Meta(
            api_version="1.0",
        ),
        files=tuple(files),
    )



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import pathlib

import aiosqlite
import pytest

from ... import errors, model
from ...components.yanking import GlobYankProvider, SqliteYankProvider


@pytest.fixture
def project_page() -> model.ProjectDetail:
    return model.ProjectDetail(
        model.Meta("1.0"),
        "project",
        (
            model.File("project-1.0.whl", "url", {}),
            model.File("project-1.0.tar.gz", "url", {}),
        ),
    )


@pytest.mark.asyncio
async def test_glob_provider__yanked_files(
    tmp_path: pathlib.Path,
    project_page: model.ProjectDetail,
) -> None:
    file = tmp_path / "yank_config.json"
    file.write_text(
        data='{"project": ["*.whl", "bad"], "pandas": ["*[!.whl]", "not supported"]}',
    )

    provider = GlobYankProvider(yank_config_file=file)

    res = await provider.yanked_files(project_page)
    assert {'project-1.0.whl': 'bad'} == res


@pytest.mark.asyncio
async def test_glob_provider__yanked_versions(
    tmp_path: pathlib.Path,
    project_page: model.ProjectDetail,
) -> None:
    file = tmp_path / "yank_config.json"
    file.write_text(
        data='{"numpy": ["*.whl", "bad"], "pandas": ["*[!.whl]", "not supported"]}',
    )

    provider = GlobYankProvider(yank_config_file=file)

    res = await provider.yanked_versions(project_page)
    assert {} == res


@pytest.mark.parametrize(
    "json_string", [
        "42", "true", '["a", "b"]', "null", '"ciao"',
    ],
)
def test_load_config_wrong_type(
    json_string: str,
    tmp_path: pathlib.PosixPath,
) -> None:
    file = tmp_path / "yank_config.json"
    file.write_text(
        data=json_string,
    )
    with pytest.raises(
        errors.InvalidConfigurationError,
        match=(
            f"Invalid configuration file. {str(file)} must contain a dictionary."
        ),
    ):
        GlobYankProvider(yank_config_file=file)


def test_load_config_wrong_format(
    tmp_path: pathlib.PosixPath,
) -> None:
    file = tmp_path / "yank_config.json"
    file.write_text(
        data='{"a": "b"}',
    )
    with pytest.raises(
        errors.InvalidConfigurationError,
        match=(
            f'Invalid yank configuration file. {str(file)} must'
            ' contain a dictionary mapping a project name to a tuple'
            ' containing a glob pattern and a yank reason.'
        ),
    ):
        GlobYankProvider(yank_config_file=file)


def test_load_config_malformed_json(
    tmp_path: pathlib.PosixPath,
) -> None:
    file = tmp_path / "yank_config.json"
    file.write_text(
        data='a',
    )
    with pytest.raises(
        errors.InvalidConfigurationError,
        match="Invalid json file",
    ):
        GlobYankProvider(yank_config_file=file)


@pytest.mark.asyncio
async def test_sqlite_provider__yanked_versions(
    tmp_path: pathlib.Path,
    project_page: model.ProjectDetail,
) -> None:
    database_path = tmp_path / "temp.db"
    async with aiosqlite.connect(database_path) as database:
        provider = SqliteYankProvider(database)
        await provider._init_db()
        await provider._database.executemany(
            "INSERT INTO yanked_versions (project_name, version, reason)"
            " VALUES(:project_name, :version, :reason)",
            [
                {"project_name": "project", "version": "1.0", "reason": "reason1"},
                {"project_name": "project", "version": "2.0", "reason": "reason2"},
            ],
        )

        await provider._database.commit()

        versions = await provider.yanked_versions(project_page)

    assert versions == {"1.0": "reason1", "2.0": "reason2"}


@pytest.mark.asyncio
async def test_sqlite_provider__yanked_files(
    tmp_path: pathlib.Path,
    project_page: model.ProjectDetail,
) -> None:
    database_path = tmp_path / "temp.db"
    async with aiosqlite.connect(database_path) as database:
        provider = SqliteYankProvider(database)
        await provider._init_db()
        await provider._database.executemany(
            "INSERT INTO yanked_releases (project_name, file_name, reason)"
            " VALUES(:project_name, :filename, :reason)",
            [
                {"project_name": "project", "filename": "project-1.0.whl", "reason": "reason1"},
                {"project_name": "project", "filename": "project-2.0.whl", "reason": "reason2"},
            ],
        )

        await provider._database.commit()

        versions = await provider.yanked_files(project_page)

    assert versions == {"project-1.0.whl": "reason1", "project-2.0.whl": "reason2"}



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import asyncio
import typing

from packaging.version import Version

from .. import errors, model
from .priority_selected import PrioritySelectedProjectsRepository


class MergedRepository(PrioritySelectedProjectsRepository):
    """
    Represents a merged view of all the given (unsorted) repositories

    NOTICE: The MergedRepository is combining the given repositories without
            giving exclusivity of a source of a specific package to any particular
            repositories. As a result, this implementation is vulnerable to
            dependency confusion. There are cases where this behaviour is desirable
            hence its existence, but if you are unsure of those reasons, consider
            using the :class:`PrioritySelectedProjectsRepository` instead.
    """
    async def get_project_page(
        self,
        project_name: str,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.ProjectDetail:
        """Retrieves a project page for the specified normalized project name
        by searching through the grouped list of sources and blending them together.
        """
        # Keep track of unique filenames for the merged files.
        files: typing.Dict[str, model.File] = {}

        results: list[BaseException | model.ProjectDetail] = await asyncio.gather(
            *(
                source.get_project_page(
                    project_name,
                    request_context=request_context,
                )
                for source in self.sources
            ),
            return_exceptions=True,
        )

        project_pages: list[model.ProjectDetail] = []
        for result in results:
            if isinstance(result, BaseException):
                if not isinstance(result, errors.PackageNotFoundError):
                    raise result
            else:
                for file in result.files:
                    # Only add the file if the filename hasn't been seen before.
                    files.setdefault(file.filename, file)
                project_pages.append(result)

        if not project_pages:
            raise errors.PackageNotFoundError(
                package_name=project_name,
            )

        # Downgrade the API version to the lowest available, as it will not be
        # possible to calculate the missing files to perform a version upgrade.
        api_version = str(
            min((
                Version(result.meta.api_version) for result in project_pages
            )),
        )

        return model.ProjectDetail(
            meta=model.Meta(api_version),
            name=project_pages[0].name,
            files=tuple(files.values()),
        )



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

class PackageNotFoundError(LookupError):
    msg_format = (
        "Package '{package_name}' was not found in the configured source"
    )

    def __init__(self, package_name: str, *args: object):
        msg = self.msg_format.format(package_name=package_name)
        super().__init__(msg, *args)


class SourceRepositoryUnavailable(Exception):
    pass


class NotNormalizedProjectName(Exception):
    pass


class UnsupportedSerialization(ValueError):
    msg_format = (
        "Unsupported format '{format_name}'."
    )

    def __init__(self, format_name: str, *args: object):
        msg = self.msg_format.format(format_name=format_name)
        super().__init__(msg, *args)


class UnsupportedAPIVersion(Exception):
    pass


class ResourceUnavailable(LookupError):
    msg_format = (
        "Resource '{resource_name}' was not found in the configured source"
    )

    def __init__(self, resource_name: str, *args: object):
        msg = self.msg_format.format(resource_name=resource_name)
        super().__init__(msg, *args)


class InvalidConfigurationError(ValueError):
    pass


class InvalidPackageError(ValueError):
    pass



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import html.parser
import typing


class HTMLElement:
    __slots__ = ("tag", "attrs", "content")

    def __init__(
            self,
            tag: str,
            attrs: dict[str, str | None],
            content: str | None = None,
    ):
        self.tag = tag
        self.attrs = attrs
        self.content = content

    def __str__(self) -> str:
        result = f"<{self.tag}"
        result += "".join(f' {key}="{value if value else ""}"' for key, value in self.attrs.items())
        result += f">{self.content}</{self.tag}>" if self.content else "/>"
        return result

    def __repr__(self) -> str:
        return f"{type(self).__name__}(tag={self.tag}, attrs={self.attrs}, content={self.content})"

    def __eq__(self, other: typing.Any) -> bool:
        if not isinstance(other, HTMLElement):
            return False
        return self.tag == other.tag and self.attrs == other.attrs and self.content == other.content


class SimpleHTMLParser(html.parser.HTMLParser):
    """HTML parser for very basic use-cases, can break with more complex HTML

    pip has an even simpler implementation, but that is not able to parse the Project List page
    (https://github.com/pypa/pip/blob/main/src/pip/_internal/index/collector.py#L395-L420_).


    Important:
        Parsing of invalid HTML will result in tag's and data being mixed up, so strange results.
        The only time when this parser seems to error is with decoding issues.
    """

    def __init__(self, *, convert_charrefs: bool = True):
        super().__init__(convert_charrefs=convert_charrefs)
        self.declaration: str | None = None
        self.elements: list[HTMLElement] = []
        self._current_tag: str | None = None
        self._current_data: str | None = None

    def handle_decl(self, decl: str) -> None:
        self.declaration = decl

    def handle_starttag(self, tag: str, attrs: list[tuple[str, str | None]]) -> None:
        self.elements.append(HTMLElement(tag, dict(attrs)))
        self._current_tag = tag
        self._current_data = None

    def handle_data(self, data: str) -> None:
        self._current_data = data

    def handle_endtag(self, tag: str) -> None:
        if tag == self._current_tag:
            self.elements[-1].content = self._current_data



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from unittest import mock

import httpx
import pytest
from pytest_httpx import HTTPXMock

from ... import errors, model
from ...components.http import HttpRepository


@pytest.mark.parametrize(
    ("text", "header"), [
        (
            '''
        <html>
            <head><title>Test Page</title></head>
            <body>
                <a href="test1.whl#hash=test_hash">test1.whl</a>
                <a href="http://test2.whl">test2.whl</a>
            </body>
        </html>
        ''', "application/vnd.pypi.simple.v1+html",
        ),
        (
            '''
        <html>
            <head><title>Test Page</title></head>
            <body>
                <a href="test1.whl#hash=test_hash">test1.whl</a>
                <a href="http://test2.whl">test2.whl</a>
            </body>
        </html>
        ''', "",
        ),
        (
            '''
        <html>
            <head><title>Test Page</title></head>
            <body>
                <a href="test1.whl#hash=test_hash">test1.whl</a>
                <a href="http://test2.whl">test2.whl</a>
            </body>
        </html>
        ''', "text/html",
        ),
        (
            '''
        {
            "meta": {
                "api-version": "1.0"
            },
            "name": "project",
            "files": [
                {
                    "filename": "test1.whl",
                    "url": "test1.whl",
                    "hashes": {"hash": "test_hash"}
                },
                {
                    "filename": "test2.whl",
                    "url": "http://test2.whl",
                    "hashes": {}
                }
            ]
        }
        ''', "application/vnd.pypi.simple.v1+json",
        ),
    ],
)
@pytest.mark.asyncio
async def test_get_project_page(text: str, header: str, httpx_mock: HTTPXMock) -> None:
    repository = HttpRepository(url="https://example.com/simple/")
    httpx_mock.add_response(content=text, headers={"content-type": header})

    resp = await repository.get_project_page("project")
    assert resp == model.ProjectDetail(
        name="project",
        meta=model.Meta(
            api_version="1.0",
        ),
        files=(
            model.File(
                filename="test1.whl",
                url="https://example.com/simple/project/test1.whl",
                hashes={"hash": "test_hash"},
            ),
            model.File(
                filename="test2.whl",
                url="http://test2.whl",
                hashes={},
            ),
        ),
    )


@pytest.mark.asyncio
async def test_get_project_page_unsupported_serialization(httpx_mock: HTTPXMock) -> None:
    repository = HttpRepository(url="https://example.com/simple/")
    httpx_mock.add_response(content="abc", headers={"content-type": "multipart/form-data"})

    with pytest.raises(errors.UnsupportedSerialization):
        await repository.get_project_page("project")


@pytest.mark.asyncio
async def test_get_project_page__package_not_found(httpx_mock: HTTPXMock) -> None:
    repository = HttpRepository(url="https://example.com/simple/")
    httpx_mock.add_response(status_code=404)

    with pytest.raises(errors.PackageNotFoundError):
        await repository.get_project_page("project")


@pytest.mark.parametrize(
    "status_code", [400, 401, 403, 500, 501],
)
@pytest.mark.asyncio
async def test_get_project_page__bad_status_code(httpx_mock: HTTPXMock, status_code: int) -> None:
    repository = HttpRepository(url="https://example.com/simple/")
    httpx_mock.add_response(status_code=status_code)

    with pytest.raises(errors.SourceRepositoryUnavailable) as exc:
        await repository.get_project_page("project")
    exc.value.__context__ == httpx.HTTPStatusError


@pytest.mark.asyncio
async def test_get_project_page__http_error(httpx_mock: HTTPXMock) -> None:
    repository = HttpRepository(url="https://example.com/simple/")
    httpx_mock.add_exception(httpx.HTTPError("Error"))

    with pytest.raises(errors.SourceRepositoryUnavailable) as exc:
        await repository.get_project_page("project")
    exc.value.__context__ == httpx.HTTPError


@pytest.mark.parametrize(
    ("text", "header"), [
        (
            '''
        <html>
            <head><title>Test Page</title></head>
            <body>
                <a href="/p1/">p1</a>
                <a href="/p2/">p2</a>
            </body>
        </html>
        ''', "text/html",
        ),
        (
            '''
        {
            "meta": {
                "api-version": "1.0"
            },
            "projects": [
                {"name": "p1"},
                {"name": "p2"}
            ]
        }
        ''', "application/vnd.pypi.simple.v1+json",
        ),
    ],
)
@pytest.mark.asyncio
async def test_get_project_list(text: str, header: str, httpx_mock: HTTPXMock) -> None:
    repository = HttpRepository(url="https://example.com/simple/")
    httpx_mock.add_response(content=text, headers={"content-type": header})

    resp = await repository.get_project_list()
    assert resp == model.ProjectList(
        meta=model.Meta(
            api_version="1.0",
        ),
        projects=frozenset([
            model.ProjectListElement(name="p1"),
            model.ProjectListElement(name="p2"),
        ]),
    )


@pytest.mark.parametrize(
    "status_code", [400, 401, 403, 404, 500, 501],
)
@pytest.mark.asyncio
async def test_get_project_list__bad_status_code(httpx_mock: HTTPXMock, status_code: int) -> None:
    repository = HttpRepository(url="https://example.com/simple/")
    httpx_mock.add_response(status_code=status_code)

    with pytest.raises(errors.SourceRepositoryUnavailable) as exc:
        await repository.get_project_list()
    exc.value.__context__ == httpx.HTTPStatusError


@pytest.mark.asyncio
async def test_get_project_list__http_error(httpx_mock: HTTPXMock) -> None:
    repository = HttpRepository(url="https://example.com/simple/")
    httpx_mock.add_exception(httpx.HTTPError("Error"))

    with pytest.raises(errors.SourceRepositoryUnavailable) as exc:
        await repository.get_project_list()
    exc.value.__context__ == httpx.HTTPError


@pytest.fixture
def project_detail() -> model.ProjectDetail:
    return model.ProjectDetail(
        meta=model.Meta("1.0"),
        name="numpy",
        files=(
            model.File(
                filename="numpy-1.0.whl",
                url="http://my_url/numpy-1.0.whl",
                hashes={},
                dist_info_metadata={"sha256": "..."},
            ),
            model.File(
                filename="numpy-2.0.whl",
                url="http://my_url/numpy-2.0.whl",
                hashes={},
            ),
        ),
    )


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "source_etag", [None, "source_etag"],
)
async def test_get_resource(
    project_detail: model.ProjectDetail,
    source_etag: str | None,
    httpx_mock: HTTPXMock,
) -> None:
    repository = HttpRepository(url="https://example.com/simple/")
    httpx_mock.add_response(headers={"ETag": source_etag} if source_etag else {})

    with mock.patch.object(repository, "get_project_page", return_value=project_detail):
        resp = await repository.get_resource("numpy", "numpy-2.0.whl")
        assert isinstance(resp, model.HttpResource)
        assert resp.url == "http://my_url/numpy-2.0.whl"
        assert resp.context.get("etag") == source_etag


@pytest.mark.asyncio
async def test_get_resource__not_modified(
    project_detail: model.ProjectDetail,
    httpx_mock: HTTPXMock,
) -> None:
    repository = HttpRepository(url="https://example.com/simple/")
    etag = "my_etag"
    httpx_mock.add_response(headers={"ETag": etag})

    with (
        mock.patch.object(repository, "get_project_page", return_value=project_detail),
        pytest.raises(model.NotModified),
    ):
        await repository.get_resource(
            project_name="numpy",
            resource_name="numpy-2.0.whl",
            request_context=model.RequestContext(repository, {"etag": etag}),
        )


@pytest.mark.asyncio
async def test_get_resource_unavailable(project_detail: model.ProjectDetail, httpx_mock: HTTPXMock) -> None:
    repository = HttpRepository(url="https://example.com/simple/")
    with mock.patch.object(repository, "get_project_page", return_value=project_detail):
        with pytest.raises(errors.ResourceUnavailable, match="numpy-3.0.whl"):
            await repository.get_resource("numpy", "numpy-3.0.whl")


@pytest.mark.asyncio
async def test_get_resource_project_unavailable(httpx_mock: HTTPXMock) -> None:
    repository = HttpRepository(url="https://example.com/simple/")
    with mock.patch.object(repository, "get_project_page", side_effect=errors.PackageNotFoundError("numpy")):
        with pytest.raises(errors.ResourceUnavailable, match="numpy-3.0.whl"):
            await repository.get_resource("numpy", "numpy-3.0.whl")


@pytest.mark.asyncio
async def test_get_resource__http_error(httpx_mock: HTTPXMock, project_detail: model.ProjectDetail) -> None:
    repository = HttpRepository(url="https://example.com/simple/")
    httpx_mock.add_exception(httpx.HTTPError("error"), method="HEAD")

    with (
        pytest.raises(errors.SourceRepositoryUnavailable) as exc,
        mock.patch.object(repository, "get_project_page", return_value=project_detail),
    ):
        await repository.get_resource("numpy", "numpy-2.0.whl")
    exc.value.__context__ == httpx.HTTPError


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "source_etag", [None, "source_etag"],
)
async def test_get_resource_metadata(
    project_detail: model.ProjectDetail,
    source_etag: str | None,
    httpx_mock: HTTPXMock,
) -> None:
    repository = HttpRepository(url="https://example.com/simple/")
    httpx_mock.add_response(headers={"ETag": source_etag} if source_etag else {})

    with mock.patch.object(repository, "get_project_page", return_value=project_detail):
        resp = await repository.get_resource("numpy", "numpy-1.0.whl.metadata")
        assert isinstance(resp, model.HttpResource)
        assert resp.url == "http://my_url/numpy-1.0.whl.metadata"
        assert resp.context.get("etag") == source_etag

        with pytest.raises(errors.ResourceUnavailable, match="numpy-2.0.whl.metadata"):
            await repository.get_resource("numpy", "numpy-2.0.whl.metadata")


@pytest.mark.asyncio
async def test_get_resource_metadata__unavailable(
    project_detail: model.ProjectDetail,
) -> None:
    repository = HttpRepository(url="https://example.com/simple/")
    with (
        mock.patch.object(repository, "get_project_page", return_value=project_detail),
        pytest.raises(errors.ResourceUnavailable, match="numpy-2.0.whl.metadata"),
    ):
        await repository.get_resource("numpy", "numpy-2.0.whl.metadata")



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import pathlib

import pytest

from ... import errors, model
from ...components.allow_listed import AllowListedRepository
from .fake_repository import FakeRepository


@pytest.mark.asyncio
async def test_special_get_project_page(tmp_path: pathlib.PosixPath) -> None:
    special_case_file = tmp_path / "special_cases.json"
    special_case_file.write_text('{"numpy": "url", "pandas": "url"}')

    repo = AllowListedRepository(
        source=FakeRepository(
            project_pages=[
                model.ProjectDetail(model.Meta("1.0"), "numpy", files=()),
                model.ProjectDetail(model.Meta("1.0"), "tensorflow", files=()),
            ],
        ),
        special_case_file=special_case_file,
    )

    resp = await repo.get_project_page("numpy")
    assert resp == model.ProjectDetail(model.Meta("1.0"), "numpy", files=())

    with pytest.raises(errors.PackageNotFoundError, match="package"):
        await repo.get_project_page("package")

    with pytest.raises(errors.PackageNotFoundError, match="tensorflow"):
        await repo.get_project_page("tensorflow")


@pytest.mark.asyncio
async def test_get_project_list(tmp_path: pathlib.PosixPath) -> None:
    special_case_file = tmp_path / "special_cases.json"
    special_case_file.write_text('{"numpy": "url", "pandas": "url"}')

    repo = AllowListedRepository(
        source=FakeRepository(),
        special_case_file=special_case_file,
    )

    res = await repo.get_project_list()

    assert res == model.ProjectList(
        meta=model.Meta("1.0"),
        projects=frozenset([model.ProjectListElement("numpy"), model.ProjectListElement("pandas")]),
    )


@pytest.mark.asyncio
async def test_get_resources(tmp_path: pathlib.PosixPath) -> None:
    special_case_file = tmp_path / "special_cases.json"
    special_case_file.write_text('{"numpy": "url", "pandas": "url"}')

    repo = AllowListedRepository(
        source=FakeRepository(
            resources={
                "gunicorn-0.7.whl": model.HttpResource(
                    url="gunicorn_url",
                ),
                "numpy-0.7.whl": model.HttpResource(
                    url="numpy_url",
                ),
            },
        ),
        special_case_file=special_case_file,
    )

    with pytest.raises(
        errors.ResourceUnavailable,
        match="gunicorn-0.7.whl",
    ):
        await repo.get_resource("gunicorn", "gunicorn-0.7.whl")

    with pytest.raises(
        errors.ResourceUnavailable,
        match="pandas.whl",
    ):
        await repo.get_resource("pandas", "pandas.whl")

    result = await repo.get_resource("numpy", "numpy-0.7.whl")

    assert isinstance(result, model.HttpResource)
    assert result.url == "numpy_url"


@pytest.mark.parametrize(
    "json_string", [
        "42", "true", '["a", "b"]', "null", '"ciao"',
    ],
)
@pytest.mark.asyncio
async def test_load_config_wrong_type(
    json_string: str,
    tmp_path: pathlib.PosixPath,
) -> None:
    file = tmp_path / "yank_config.json"
    file.write_text(
        data=json_string,
    )
    with pytest.raises(
        errors.InvalidConfigurationError,
        match=(
            f"Invalid configuration file. {str(file)} must contain a dictionary."
        ),
    ):
        AllowListedRepository(
            source=FakeRepository(),
            special_case_file=file,
        )


def test_load_config_wrong_format(
    tmp_path: pathlib.PosixPath,
) -> None:
    file = tmp_path / "yank_config.json"
    file.write_text(
        data='{"a": ["b", "c"]}',
    )
    with pytest.raises(
        errors.InvalidConfigurationError,
        match=(
            f'Invalid special case configuration file. {str(file)} '
            'must contain a dictionary mapping a project name to a tuple'
            ' containing a glob pattern and a yank reason.'
        ),
    ):
        AllowListedRepository(
            source=FakeRepository(),
            special_case_file=file,
        )


@pytest.mark.asyncio
async def test_load_config_malformed_json(
    tmp_path: pathlib.PosixPath,
) -> None:
    file = tmp_path / "yank_config.json"
    file.write_text(
        data='a',
    )
    with pytest.raises(
        errors.InvalidConfigurationError,
        match="Invalid json file",
    ):
        AllowListedRepository(
            source=FakeRepository(),
            special_case_file=file,
        )



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from datetime import datetime
import hashlib
import os
import pathlib

from packaging.utils import canonicalize_name

from .. import errors, model
from .core import SimpleRepository


class LocalRepository(SimpleRepository):
    """
    Creates a simple repository from a local directory.
    The directory must contain a subdirectory for each project,
    named as the normalized project name. Each subdirectory will
    contain the distributions associated with that project.
    Each file in a project page is mapped to a URL with the
    following structure: file:// index_path / project_name / file_name.
    """
    def __init__(
        self,
        index_path: pathlib.Path,
    ) -> None:
        if not index_path.is_dir():
            raise ValueError("index_path must be a directory")
        self._index_path = index_path.resolve()

    async def get_project_list(
        self,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.ProjectList:
        return model.ProjectList(
            meta=model.Meta("1.0"),
            projects=frozenset(
                model.ProjectListElement(x.name)
                for x in self._index_path.iterdir()
                if x.is_dir() and x.name == canonicalize_name(x.name)
            ),
        )

    async def get_project_page(
        self,
        project_name: str,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.ProjectDetail:
        project_dir = (self._index_path / project_name).resolve()
        if not project_dir.is_dir():
            raise errors.PackageNotFoundError(project_name)

        files = []
        for file in sorted(project_dir.iterdir()):
            if not file.is_file():
                continue
            file_stat = os.stat(file)
            files.append(
                model.File(
                    filename=file.name,
                    url=f"file://{file.resolve()}",
                    hashes={},
                    upload_time=datetime.utcfromtimestamp(
                        file_stat.st_mtime,
                    ),
                    size=file_stat.st_size,
                ),
            )

        return model.ProjectDetail(
            meta=model.Meta("1.1"),
            name=project_name,
            files=tuple(files),
        )

    async def get_resource(
        self,
        project_name: str,
        resource_name: str,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.Resource:
        repository_uri = (self._index_path / project_name).resolve()
        resource_uri = (repository_uri / resource_name).resolve()

        if (
            not repository_uri.is_relative_to(self._index_path) or
            not resource_uri.is_relative_to(repository_uri)
        ):
            raise ValueError(
                f"{resource_uri} is not contained in {repository_uri}",
            )
        if not resource_uri.is_file():
            raise errors.ResourceUnavailable(resource_name)

        # Calculating on the fly the hash of the whole package can be too slow.
        # "mtime + size" provide a good approximation to detect if the package has been changed.
        etag_base = str(resource_uri.stat().st_mtime) + "-" + str(resource_uri.stat().st_size)
        etag = hashlib.md5(etag_base.encode(), usedforsecurity=False).hexdigest()

        return model.LocalResource(
            path=resource_uri,
            to_cache=False,  # Prevent caching of locally stored resources.
            context=model.Context(etag=etag),
        )



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import pathlib
import typing
from unittest import mock
import zipfile

import httpx
import pytest

from ... import errors, model
from ...components.core import SimpleRepository
from ...components.metadata_injector import MetadataInjectorRepository
from .fake_repository import FakeRepository


@pytest.fixture
def repository() -> MetadataInjectorRepository:
    return MetadataInjectorRepository(
        source=FakeRepository(
            project_pages=[
                model.ProjectDetail(
                    model.Meta("1.0"), "numpy", files=(
                        model.File("numpy-1.0-any.whl", "url", {}),
                        model.File("numpy-1.0.tar.gz", "url", {}),
                    ),
                ),
            ],
            resources={
                "numpy-1.0-any.whl": model.HttpResource(
                    url="numpy_url",
                ),
                "numpy-1.0.tar.gz": model.HttpResource(
                    url="numpy_url",
                ),
                "numpy-2.0-any.whl": model.LocalResource(
                    path=pathlib.Path("file/path"),
                ),
            },
        ),
        http_client=mock.AsyncMock(),
    )


def test_add_metadata_attribute(repository: MetadataInjectorRepository) -> None:
    project_page = model.ProjectDetail(
       model.Meta("1.0"),
       "numpy",
       (
            model.File("numpy-1.0-any.whl", "/numpy-1.0-any.whl", {}, dist_info_metadata=None),
            model.File("numpy-1.0-any.whl", "/numpy-1.0-any.whl", {}, dist_info_metadata=False),
            model.File("numpy-1.0-any.tar.gz", "/numpy-1.0-any.tar.gz", {}),
            model.File(
                "numpy-1.1-any.whl", "/numpy-1.0-any.whl", {}, dist_info_metadata={"sha": "..."},
            ),
       ),
    )
    result = repository._add_metadata_attribute(project_page)

    assert result.files[0].dist_info_metadata is True
    assert result.files[1].dist_info_metadata is True
    assert result.files[2].dist_info_metadata is None
    assert result.files[3].dist_info_metadata == {"sha": "..."}


@pytest.mark.parametrize(
    "namelist, metadata_name", [
        (
            [
                "my_package/files",
                "not_my_package-0.0.1.dist-info/METADATA",
                "my_package-0.0.1.dist-info/METADATA",
            ], "my_package-0.0.1.dist-info/METADATA",
        ),
        (
            [
                "my_package/files",
                "my_package-0.0.1.dist-info/METADATA",
                "My_Package-0.0.1.dist-info/METADATA",
            ], "my_package-0.0.1.dist-info/METADATA",
        ),
        (
            [
                "my_package/files",
                "My_Package-0.0.1.dist-info/METADATA",
            ], "My_Package-0.0.1.dist-info/METADATA",
        ),
        (
            [
                "my_package/files",
                "my.package-0.0.1.dist-info/METADATA",
            ], "my.package-0.0.1.dist-info/METADATA",
        ),
        (
            [
                "my_package/files",
                "my-package-0.0.1.dist-info/METADATA",
            ], "my-package-0.0.1.dist-info/METADATA",
        ),
    ],
)
def test_get_metadata_from_package(repository: MetadataInjectorRepository, namelist: list[str], metadata_name: str) -> None:
    ziparchive = mock.MagicMock(spec=zipfile.ZipFile)
    ziparchive_ctx = ziparchive.__enter__.return_value
    read_method = ziparchive_ctx.read
    ziparchive_ctx.namelist.return_value = namelist

    with mock.patch('zipfile.ZipFile', return_value=ziparchive):
        repository._get_metadata_from_package(pathlib.Path('my_package') / 'my_package-0.0.1-anylinux.whl')

        read_method.assert_called_once_with(metadata_name)


def test_get_metadata_from_package__not_wheel(repository: MetadataInjectorRepository) -> None:
    with pytest.raises(ValueError, match="Package provided is not a wheel"):
        repository._get_metadata_from_package(pathlib.Path('my_package') / 'package.tar.gz')


@pytest.mark.parametrize(
    "namelist", [
        [
            "my_package/files",
            "not_my_package-0.0.1.dist-info/METADATA",
        ],
        [
            "my_package/files",
            "my_package.dist-info/METADATA",
        ],
        [
            "my_package/files",
            "my_package-0.0.1.dist-info/NOT_METADATA",
        ],
    ],
)
def test_get_metadata_from_package__missing_metadata(repository: MetadataInjectorRepository, namelist: list[str]) -> None:
    m_zipfile_cls = mock.MagicMock(spec=zipfile.ZipFile)
    m_zipfile_cls.return_value.__enter__.return_value.namelist.return_value = [
        "not_my_package-0.0.1.dist-info/METADATA",
    ]

    with mock.patch('zipfile.ZipFile', spec=zipfile.ZipFile, new=m_zipfile_cls):
        with pytest.raises(
            errors.InvalidPackageError,
            match="Provided wheel doesn't contain a metadata file.",
        ):
            repository._get_metadata_from_package(
                package_path=pathlib.Path('my_package') / 'my_package-0.0.1-anylinux.whl',
            )


@pytest.mark.asyncio
async def test_get_project_page(
    repository: MetadataInjectorRepository,
) -> None:
    result = await repository.get_project_page("numpy")
    assert result.files[0].dist_info_metadata is True


@pytest.mark.asyncio
async def test_get_resource(
    repository: MetadataInjectorRepository,
) -> None:
    with mock.patch.object(
        repository,
        "_download_metadata",
        return_value="downloaded_meta",
    ):
        response = await repository.get_resource("numpy", "numpy-1.0-any.whl.metadata")

    assert isinstance(response, model.TextResource)
    assert response.text == "downloaded_meta"


@pytest.mark.asyncio
async def test_get_resource__local_resource(
    repository: MetadataInjectorRepository,
) -> None:
    with mock.patch.object(
        repository,
        "_get_metadata_from_package",
        return_value="downloaded_meta",
    ):
        response = await repository.get_resource("numpy", "numpy-2.0-any.whl.metadata")

    assert isinstance(response, model.TextResource)
    assert response.text == "downloaded_meta"


@pytest.mark.asyncio
async def test_get_resource__not_valid_resource() -> None:
    source_repo = mock.Mock(spec=SimpleRepository)
    source_repo.get_resource.side_effect = [
        errors.ResourceUnavailable('name'),
        model.TextResource(text='/etc/passwd'),
    ]
    repo = MetadataInjectorRepository(
        source=typing.cast(SimpleRepository, source_repo),
        http_client=mock.AsyncMock(),
    )
    with pytest.raises(errors.ResourceUnavailable, match='Unable to fetch the resource needed to extract the metadata'):
        await repo.get_resource("numpy", "numpy-1.0-any.whl.metadata")


@pytest.mark.parametrize(
    "resource_name", ["numpy-1.0-any.whl", "numpy-1.0.tar.gz"],
)
@pytest.mark.asyncio
async def test_get_resource__not_metadata(
    repository: MetadataInjectorRepository,
    resource_name: str,
) -> None:
    response = await repository.get_resource("numpy", resource_name)
    assert isinstance(response, model.HttpResource)
    assert response.url == "numpy_url"


@pytest.mark.asyncio
async def test_download_metadata(
    repository: MetadataInjectorRepository,
) -> None:
    with (
        mock.patch.object(
            repository,
            "_get_metadata_from_package",
        ) as get_metadata_from_package_mock,
        mock.patch("simple_repository.utils.download_file") as download_file_mock,
    ):
        await repository._download_metadata("name", "url", httpx.AsyncClient())

    get_metadata_from_package_mock.assert_called_once()
    download_file_mock.assert_awaited_once()



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import pytest

from ... import model
from ...components.yanking import YankProvider, YankRepository
from .fake_repository import FakeRepository


class FakeYankProvider(YankProvider):
    async def yanked_versions(self, project_page: model.ProjectDetail) -> dict[str, str]:
        return {"1.0": "reason"}

    async def yanked_files(self, project_page: model.ProjectDetail) -> dict[str, str]:
        return {"project-1.0-any.whl": "reason"}


@pytest.fixture
def project_page() -> model.ProjectDetail:
    return model.ProjectDetail(
        model.Meta("1.0"), name="project", files=(
            model.File("project-1.0-any.whl", "url", {}),
            model.File("project-1.0.tar.gz", "url", {}),
            model.File("project-1.1.tar.gz", "url", {}),
        ),
    )


@pytest.fixture
def repository(project_page: model.ProjectDetail) -> YankRepository:
    source = FakeRepository(
        project_pages=[project_page],
    )
    provider = FakeYankProvider()

    return YankRepository(
        source=source,
        yank_provider=provider,
    )


@pytest.mark.asyncio
async def test_get_project_page(repository: YankRepository) -> None:
    result = await repository.get_project_page("project")

    assert result.files[0].yanked == "reason"
    assert result.files[1].yanked == "reason"
    assert result.files[2].yanked is None



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import pytest

from ...components.core import SimpleRepositoryMeta
from ...model import RequestContext
from .fake_repository import FakeRepository


class TestClass(metaclass=SimpleRepositoryMeta):
    async def get_project_page(self, *, request_context: RequestContext = RequestContext.DEFAULT) -> RequestContext:
        return request_context

    async def get_project_list(self, *, request_context: RequestContext = RequestContext.DEFAULT) -> RequestContext:
        return request_context

    async def get_resource(self, *, request_context: RequestContext = RequestContext.DEFAULT) -> RequestContext:
        return request_context


@pytest.mark.asyncio
async def test_decorated_get_project_page__default_context() -> None:
    test_object = TestClass()
    context = await test_object.get_project_page()
    assert context != RequestContext.DEFAULT
    assert isinstance(context, RequestContext)
    assert context.repository is test_object  # type: ignore[comparison-overlap]


@pytest.mark.asyncio
async def test_decorated_get_project_page__passed_context() -> None:
    test_object = TestClass()
    request_context = RequestContext(
        repository=FakeRepository(),
    )
    context = await test_object.get_project_page(request_context=request_context)
    assert context is request_context


@pytest.mark.asyncio
async def test_decorated_get_project_list__default_context() -> None:
    test_object = TestClass()
    context = await test_object.get_project_list()
    assert context != RequestContext.DEFAULT
    assert context.repository is test_object  # type: ignore[comparison-overlap]


@pytest.mark.asyncio
async def test_decorated_get_project_list__passed_context() -> None:
    test_object = TestClass()
    request_context = RequestContext(
        repository=FakeRepository(),
    )
    context = await test_object.get_project_list(request_context=request_context)
    assert context is request_context


@pytest.mark.asyncio
async def test_decorated_get_resource__default_context() -> None:
    test_object = TestClass()
    context = await test_object.get_resource()
    assert context != RequestContext.DEFAULT
    assert isinstance(context, RequestContext)
    assert context.repository is test_object  # type: ignore[comparison-overlap]


@pytest.mark.asyncio
async def test_decorated_get_resource__passed_context() -> None:
    test_object = TestClass()
    request_context = RequestContext(
        repository=FakeRepository(),
    )
    context = await test_object.get_resource(request_context=request_context)
    assert context is request_context



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import functools
import typing

from .. import model


class SimpleRepositoryMeta(type):
    # A metaclass that swaps model.RequestContext.DEFAULT arguments for RequestContext
    # instances containing the repository upon which the method is called.
    def __new__(
            cls: typing.Type[type],
            name: str,
            bases: tuple[typing.Type[type]],
            namespace: dict[str, typing.Any],
    ) -> typing.Type[type]:
        wrapped_fn: typing.TypeAlias = typing.Callable[[typing.Any], typing.Any]

        def dec(fn: wrapped_fn) -> wrapped_fn:
            @functools.wraps(fn)
            async def wrapper(
                    self: typing.Any, *args: typing.Any, **kwargs: typing.Any,
            ) -> typing.Any:
                # If we have the default RequestContext (which is None), swap it for
                # a new context which contains self.
                if kwargs.get('request_context') is model.RequestContext.DEFAULT:
                    kwargs['request_context'] = model.RequestContext(self)
                return await fn(self, *args, **kwargs)
            return wrapper

        if 'get_project_page' in namespace:
            namespace['get_project_page'] = dec(namespace['get_project_page'])
        if 'get_project_list' in namespace:
            namespace['get_project_list'] = dec(namespace['get_project_list'])
        if 'get_resource' in namespace:
            namespace['get_resource'] = dec(namespace['get_resource'])

        result = type.__new__(cls, name, bases, dict(namespace))
        return result


class SimpleRepository(metaclass=SimpleRepositoryMeta):
    async def get_project_page(
        self,
        project_name: str,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.ProjectDetail:
        raise NotImplementedError()

    async def get_project_list(
        self,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.ProjectList:
        raise NotImplementedError()

    async def get_resource(
        self,
        project_name: str,
        resource_name: str,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.Resource:
        raise NotImplementedError()


class RepositoryContainer(SimpleRepository):
    """A base class for components that enhance the functionality of a source
    `SimpleRepository`. If not overridden, the methods provided by this class
    will delegate to the corresponding methods of the source repository.
    """
    def __init__(self, source: SimpleRepository) -> None:
        self.source = source

    async def get_project_page(
        self,
        project_name: str,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.ProjectDetail:
        return await self.source.get_project_page(project_name, request_context=request_context)

    async def get_project_list(
        self,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.ProjectList:
        return await self.source.get_project_list(request_context=request_context)

    async def get_resource(
        self,
        project_name: str,
        resource_name: str,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.Resource:
        return await self.source.get_resource(
            project_name,
            resource_name,
            request_context=request_context,
        )



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

"""
Specialised PEP-503 simple repository components exposed through a standardised
client interface, suitable for re-use in both client and server implementations
"""

from ._version import version as __version__  # noqa
from .components.core import SimpleRepository  # noqa

__all__ = ['__version__', 'SimpleRepository']



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from typing import Optional

from ... import errors, model
from ...components.core import SimpleRepository


class FakeRepository(SimpleRepository):
    def __init__(
        self,
        project_list: model.ProjectList = model.ProjectList(model.Meta('1.0'), frozenset()),
        project_pages: Optional[list[model.ProjectDetail]] = None,
        resources: Optional[dict[str, model.Resource]] = None,
    ) -> None:
        self.project_list = project_list
        if project_pages:
            self.project_pages = {
                project.name: project for project in project_pages
            }
        else:
            self.project_pages = {}
        self.resources = resources or {}

    async def get_project_page(
        self,
        project_name: str,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.ProjectDetail:
        if project_name in self.project_pages:
            return self.project_pages[project_name]
        raise errors.PackageNotFoundError(project_name)

    async def get_project_list(
        self,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.ProjectList:
        return self.project_list

    async def get_resource(
        self,
        project_name: str,
        resource_name: str,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.Resource:
        if resource := self.resources.get(resource_name):
            return resource
        raise errors.ResourceUnavailable(resource_name)



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import pytest

from .. import errors
from ..content_negotiation import Format, select_response_format


@pytest.mark.parametrize(
        "content_type, format", [
            ("", Format.HTML_LEGACY),
            ("application/vnd.pypi.simple.v1+html;q=0.8,application/vnd.pypi.simple.v1+json;q=0.9", Format.JSON_V1),
            ("text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8", Format.HTML_LEGACY),
            ("application/vnd.pypi.simple.v1+html,application/xhtml+xml;q=0.1", Format.HTML_V1),
            ("application/vnd.pypi.simple.v1+json; q = 0.9, application/vnd.pypi.simple.v1+html; q = 0.8", Format.JSON_V1),
        ],
)
def test_select_response_format(content_type: str, format: Format) -> None:
    assert select_response_format(content_type) == format


@pytest.mark.parametrize(
        "content_type", [
            "application/json",
            "application/vnd.pypi.simple.v2+json;q=0.9 , application/vnd.pypi.simple.v2+html;q=0.8",
        ],
)
def test_select_response_format_unsupported(content_type: str) -> None:
    with pytest.raises(errors.UnsupportedSerialization):
        select_response_format(content_type)



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import pytest

from ... import errors, model
from ...components.merged import MergedRepository
from .fake_repository import FakeRepository


@pytest.mark.asyncio
async def test_get_project_page() -> None:
    repo = MergedRepository(
        [
            FakeRepository(),
            FakeRepository(
                project_pages=[
                    model.ProjectDetail(
                        model.Meta('1.1'),
                        "numpy",
                        files=(
                            model.File("numpy-1.1.whl", "url1", {}, size=1),
                            model.File("numpy-1.2.whl", "url1", {}, size=1),
                        ),
                    ),
                ],
            ),
            FakeRepository(
                project_pages=[
                    model.ProjectDetail(
                        model.Meta('1.0'),
                        "numpy",
                        files=(
                            model.File("numpy-1.1.whl", "url2", {}),
                            model.File("numpy-1.3.whl", "url2", {}),
                        ),
                    ),
                ],
            ),
        ],
    )

    resp = await repo.get_project_page("numpy")

    assert resp == model.ProjectDetail(
        model.Meta('1.0'),
        "numpy",
        files=(
            model.File("numpy-1.1.whl", "url1", {}, size=1),
            model.File("numpy-1.2.whl", "url1", {}, size=1),
            model.File("numpy-1.3.whl", "url2", {}),
        ),
    )


@pytest.mark.asyncio
async def test_get_project_page_failed() -> None:
    repo = MergedRepository([
        FakeRepository() for _ in range(3)
    ])

    with pytest.raises(
        errors.PackageNotFoundError,
        match="Package 'numpy' was not found in the configured source",
    ):
        await repo.get_project_page("numpy")



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from datetime import datetime
from typing import Optional, Union

import pytest

from .. import model, parser


def test_parse_json_project_page() -> None:
    page = '''
    {
        "meta": {
            "api-version": "1.0"
        },
        "name": "holygrail",
        "files": [
            {
                "filename": "holygrail-1.0.tar.gz",
                "url": "https://example.com/files/holygrail-1.0.tar.gz",
                "hashes": {"sha256": "...", "blake2b": "..."},
                "requires-python": ">=3.7",
                "yanked": "Had a vulnerability"
            },
            {
                "filename": "holygrail-1.0-py3-none-any.whl",
                "url": "holygrail-1.0-py3-none-any.whl",
                "hashes": {},
                "core-metadata": true
            },
            {
                "filename": "holygrail-1.1-py3-none-any.whl",
                "url": "holygrail-1.0-py3-none-any.whl",
                "hashes": {},
                "core-metadata": {"sha256": "..."},
                "yanked": true
            },
            {
                "filename": "holygrail-1.2-py3-none-any.whl",
                "url": "http://unnormalized url.whl",
                "hashes": {}
            }
        ]

    }'''

    result = parser.parse_json_project_page(page)

    assert result == model.ProjectDetail(
        model.Meta("1.0"),
        "holygrail",
        (
            model.File(
                filename="holygrail-1.0.tar.gz",
                url="https://example.com/files/holygrail-1.0.tar.gz",
                hashes={"sha256": "...", "blake2b": "..."},
                requires_python=">=3.7",
                yanked="Had a vulnerability",
            ),
            model.File(
                filename="holygrail-1.0-py3-none-any.whl",
                url="holygrail-1.0-py3-none-any.whl",
                hashes={},
                dist_info_metadata=True,
            ),
            model.File(
                filename="holygrail-1.1-py3-none-any.whl",
                url="holygrail-1.0-py3-none-any.whl",
                hashes={},
                dist_info_metadata={"sha256": "..."},
                yanked=True,
            ),
            model.File(
                filename="holygrail-1.2-py3-none-any.whl",
                url="http://unnormalized%20url.whl",
                hashes={},
            ),
        ),
    )


def test_parse_html_project_page() -> None:
    page = '''
        <a href="holygrail-1.0.tar.gz#sha256=..."
            data-requires-python="&gt;=3.7"
        >holygrail-1.0.tar.gz</a>

        <a href="holygrail-1.1-py3-none-any.whl"
        >holygrail-1.1-py3-none-any.whl</a>

        <a href="unnormalized url.whl"
        >holygrail-1.2-py3-none-any.whl</a>

        <a>bad-project.whl</a>
        <a href="bad-project.whl></a>
    '''

    result = parser.parse_html_project_page(page, "holygrail")

    assert result == model.ProjectDetail(
        model.Meta("1.0"),
        "holygrail",
        (
            model.File(
                filename="holygrail-1.0.tar.gz",
                url="holygrail-1.0.tar.gz",
                hashes={"sha256": "..."},
                requires_python=">=3.7",
            ),
            model.File(
                filename="holygrail-1.1-py3-none-any.whl",
                url="holygrail-1.1-py3-none-any.whl",
                hashes={},
            ),
            model.File(
                filename="holygrail-1.2-py3-none-any.whl",
                url="unnormalized%20url.whl",
                hashes={},
            ),
        ),
    )


@pytest.mark.parametrize(
    "fragment_attr, hashes",
    [
        ('', {}),
        ('#', {}),
        ('#a=2', {'a': '2'}),
        ('#a=2&b', {'a': '2&b'}),
        ('#argh!', {}),
    ],
)
def test_parse_html_project_page_URL_fragment(
    fragment_attr: str,
    hashes: dict[str, str],
) -> None:
    page = f'''<a href="holygrail-1.0.tar.gz{fragment_attr}">holygrail-1.0.tar.gz</a>'''

    result = parser.parse_html_project_page(page, "holygrail")

    assert result == model.ProjectDetail(
        model.Meta("1.0"),
        "holygrail",
        (
            model.File(
                filename="holygrail-1.0.tar.gz",
                url="holygrail-1.0.tar.gz",
                hashes=hashes,
            ),
        ),
    )


@pytest.mark.parametrize(
    "yank_attr, yank_value",
    [
        ('data-yanked', True),
        ('data-yanked=""', True),
        ('data-yanked="reason"', "reason"),
        ('data-yanked="false"', "false"),
        ('', None),
    ],
)
def test_parse_html_project_page_yank(
    yank_attr: str,
    yank_value: Optional[Union[bool, str]],
) -> None:
    page = f'''
        <a href="holygrail-1.0.tar.gz"
            {yank_attr}
        >holygrail-1.0.tar.gz</a>
    '''

    result = parser.parse_html_project_page(page, "holygrail")

    assert result == model.ProjectDetail(
        model.Meta("1.0"),
        "holygrail",
        (
            model.File(
                filename="holygrail-1.0.tar.gz",
                url="holygrail-1.0.tar.gz",
                hashes={},
                yanked=yank_value,
            ),
        ),
    )


@pytest.mark.parametrize(
    "metadata_attr, metadata_value",
    [
        ('data-dist-info-metadata', None),
        ('data-dist-info-metadata="true"', None),
        ('data-dist-info-metadata="something incompatible"', None),
        ('data-dist-info-metadata="sha=..."', None),
        ('data-core-metadata', True),
        ('data-core-metadata="true"', True),
        ('data-core-metadata="something incompatible"', True),
        ('data-core-metadata="sha=..."', {"sha": "..."}),
        ('data-core-metadata="sha=..." data-dist-info-metadata="true"', {"sha": "..."}),
        ('', None),
    ],
)
def test_parse_html_project_page_metadata(
    metadata_attr: str,
    metadata_value: Optional[Union[bool, dict[str, str]]],
) -> None:
    page = f'''
        <a href="holygrail-1.0.tar.gz"
            {metadata_attr}
        >holygrail-1.0.tar.gz</a>
    '''

    result = parser.parse_html_project_page(page, "holygrail")

    assert result == model.ProjectDetail(
        model.Meta("1.0"),
        "holygrail",
        (
            model.File(
                filename="holygrail-1.0.tar.gz",
                url="holygrail-1.0.tar.gz",
                hashes={},
                dist_info_metadata=metadata_value,
            ),
        ),
    )


@pytest.mark.parametrize(
    "gpg_attr, gpg_value",
    [
        ('data-gpg-sig="true"', True),
        ('data-gpg-sig="false"', False),
        ('', None),
        ('data-gpg-sig="invalid"', None),
    ],
)
def test_parse_html_project_page_gpg(gpg_attr: str, gpg_value: Optional[bool]) -> None:
    page = f'''
        <a href="holygrail-1.0.tar.gz"
            {gpg_attr}
        >holygrail-1.0.tar.gz</a>
    '''

    result = parser.parse_html_project_page(page, "holygrail")

    assert result == model.ProjectDetail(
        model.Meta("1.0"),
        "holygrail",
        (
            model.File(
                filename="holygrail-1.0.tar.gz",
                url="holygrail-1.0.tar.gz",
                hashes={},
                gpg_sig=gpg_value,
            ),
        ),
    )


def test_parse_json_project_list() -> None:
    page = '''
    {
        "meta": {
            "api-version": "1.0"
        },
        "projects": [
            {
                "name": "gym"
            },
            {
                "name": "my_project"
            }
        ]
    }'''

    result = parser.parse_json_project_list(page)

    assert result == model.ProjectList(
        model.Meta("1.0"),
        frozenset([
            model.ProjectListElement("gym"),
            model.ProjectListElement("my_project"),
        ]),
    )


def test_parse_html_project_list() -> None:
    page = '''
        <a href="url">gym</a>
        <a href="url">my_project</a>
    '''

    result = parser.parse_html_project_list(page)

    assert result == model.ProjectList(
        model.Meta("1.0"),
        frozenset([
            model.ProjectListElement("gym"),
            model.ProjectListElement("my_project"),
        ]),
    )


def test_parse_json_project_page__v_1_1() -> None:
    page = '''
    {
        "meta": {
            "api-version": "1.1"
        },
        "name": "holygrail",
        "files": [
            {
                "filename": "holygrail-1.0-py3-none-any.whl",
                "url": "http://url.whl",
                "hashes": {},
                "size": 1,
                "upload-time": "2000-01-01T00:00:00.000000Z"
            },
            {
                "filename": "holygrail-1.1-py3-none-any.whl",
                "url": "http://url.whl",
                "hashes": {},
                "size": 1,
                "upload-time": "2000-01-01T00:00:00Z"
            }
        ]
    }'''

    result = parser.parse_json_project_page(page)

    assert result == model.ProjectDetail(
        model.Meta("1.1"),
        "holygrail",
        (
            model.File(
                filename="holygrail-1.0-py3-none-any.whl",
                url="http://url.whl",
                hashes={},
                size=1,
                upload_time=datetime(2000, 1, 1, 0, 0, 0),
            ),
            model.File(
                filename="holygrail-1.1-py3-none-any.whl",
                url="http://url.whl",
                hashes={},
                size=1,
                upload_time=datetime(2000, 1, 1, 0, 0, 0),
            ),
        ),
    )



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import pytest

from .. import model as m


def test_ProjectListElement__normalized_name() -> None:
    prj = m.ProjectListElement('some-.name')
    assert prj.normalized_name == 'some-name'


def test_ProjectDetail__normalized_name() -> None:
    project_detail = m.ProjectDetail(
        meta=m.Meta("1.0"),
        name="Project",
        files=(),
    )
    assert project_detail._normalized_name == "project"


def test_ProjectDetail__failed_post_init() -> None:
    with pytest.raises(
        ValueError,
        match="SimpleAPI>=1.1 requires the size field to be set for all the files.",
    ):
        m.ProjectDetail(
            meta=m.Meta("1.1"),
            name="pippo",
            files=(
                m.File(
                    filename="pippozzo",
                    url="url",
                    hashes={},
                ),
            ),
        )


def test_ProjectDetail__post_init_v1() -> None:
    project_detail = m.ProjectDetail(
        meta=m.Meta("1.0"),
        name="pippo",
        files=(
            m.File(
                filename="pippozzo",
                url="url",
                hashes={},
            ),
        ),
    )
    assert project_detail.versions is None


def test_ProjectDetail__post_init_v1_1() -> None:
    project_detail = m.ProjectDetail(
        meta=m.Meta("1.1"),
        name="pippo",
        files=(
            m.File(
                filename="pippo-1.0.tar.gz",
                url="url",
                hashes={},
                size=1,
            ),
            m.File(
                filename="pippo-2.0-anylinux-py3.whl",
                url="url",
                hashes={},
                size=1,
            ),
        ),
    )
    assert project_detail.versions == {
        "1.0", "2.0",
    }



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from dataclasses import replace
from datetime import datetime, timedelta

from .. import model
from .core import RepositoryContainer, SimpleRepository


class NewReleasesRemover(RepositoryContainer):
    """
    A component used to remove newly released projects from the source repository.
    This component can be used only if the source repository exposes the upload
    date according to PEP-700: https://peps.python.org/pep-0700/.
    """
    def __init__(
        self,
        source: SimpleRepository,
        quarantine_time: timedelta = timedelta(days=2),
        whitelist: tuple[str, ...] = tuple(),
    ) -> None:
        self._quarantine_time = quarantine_time
        self._whitelist = whitelist
        super().__init__(source)

    async def get_project_page(
        self,
        project_name: str,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.ProjectDetail:
        project_page = await super().get_project_page(
            project_name,
            request_context=request_context,
        )

        if project_name in self._whitelist:
            return project_page

        return self._exclude_recent_distributions(
            project_page=project_page,
            now=datetime.now(),
        )

    def _exclude_recent_distributions(
        self,
        project_page: model.ProjectDetail,
        now: datetime,
    ) -> model.ProjectDetail:
        filtered_files = tuple(
            file for file in project_page.files
            if not file.upload_time or
            (now - file.upload_time).total_seconds() >= self._quarantine_time.total_seconds()
        )
        return replace(project_page, files=filtered_files)



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from unittest import mock

import pytest

from ... import errors, model
from ...components.priority_selected import PrioritySelectedProjectsRepository
from .fake_repository import FakeRepository


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "version", [
        "1.0", "1.1",
    ],
)
async def test_get_project_page(version: str) -> None:
    group_repository = PrioritySelectedProjectsRepository([
        FakeRepository(),
        FakeRepository(
            project_pages=[
                model.ProjectDetail(
                    model.Meta(version),
                    "numpy",
                    files=(model.File("1", "1", {}, size=1), model.File("2", "2", {}, size=1)),
                ),
            ],
        ),
        FakeRepository(
            project_pages=[
                model.ProjectDetail(
                    model.Meta('1.0'),
                    "numpy",
                    files=(model.File("3", "3", {}),),
                ),
            ],
        ),
    ])

    resp = await group_repository.get_project_page(
        project_name="numpy",
        request_context=model.RequestContext(group_repository),
    )

    assert resp == model.ProjectDetail(
        model.Meta(version),
        "numpy",
        files=(
            model.File("1", "1", {}, size=1), model.File("2", "2", {}, size=1),
        ),
    )


@pytest.mark.asyncio
async def test_get_project_page_failed() -> None:
    group_repository = PrioritySelectedProjectsRepository([
        FakeRepository() for _ in range(3)
    ])
    with pytest.raises(
        expected_exception=errors.PackageNotFoundError,
        match=r"Package 'numpy' was not found in the configured source",
    ):
        await group_repository.get_project_page("numpy")


@pytest.mark.asyncio
async def test_blended_get_project_list() -> None:
    meta = model.Meta(api_version="1.0")
    projects_elements = [
        frozenset([
            model.ProjectListElement("a_"),
            model.ProjectListElement("c"),
        ]), frozenset([
            model.ProjectListElement("a-"),
            model.ProjectListElement("b"),
        ]), frozenset([
            model.ProjectListElement("d"),
        ]),
    ]

    group_repository = PrioritySelectedProjectsRepository(
        sources=[
            FakeRepository(
                model.ProjectList(model.Meta("1.0"), p),
            ) for p in projects_elements
        ],
    )

    result = await group_repository.get_project_list()
    # We expect only normalized results, and no duplicates.
    assert result == model.ProjectList(
        meta=meta,
        projects=frozenset([
            model.ProjectListElement("a-"),
            model.ProjectListElement("b"),
            model.ProjectListElement("c"),
            model.ProjectListElement("d"),
        ]),
    )


@pytest.mark.asyncio
async def test_blended_get_project_list_failed() -> None:
    repo = PrioritySelectedProjectsRepository(
        sources=[
            mock.AsyncMock() for _ in range(3)
        ],
    )
    assert isinstance(repo.sources[2], mock.Mock)
    repo.sources[2].get_project_list.side_effect = errors.SourceRepositoryUnavailable
    with pytest.raises(errors.SourceRepositoryUnavailable):
        await repo.get_project_list()


@pytest.mark.asyncio
async def test_blended_get_project_page_failed() -> None:
    repo = PrioritySelectedProjectsRepository(
        sources=[
            FakeRepository(
                project_pages=[
                    model.ProjectDetail(
                        meta=model.Meta("1.0"),
                        name="numpy",
                        files=(),
                    ),
                ],
            ),
            mock.AsyncMock(),
        ],
    )
    assert isinstance(repo.sources[1], mock.Mock)
    repo.sources[1].get_project_list.side_effect = Exception

    res = await repo.get_project_page("numpy")

    assert res == model.ProjectDetail(
        model.Meta("1.0"), name="numpy", files=(),
    )


def test_group_repository_failed_init() -> None:
    with pytest.raises(ValueError):
        PrioritySelectedProjectsRepository([])
    with pytest.raises(ValueError):
        PrioritySelectedProjectsRepository([FakeRepository()])


@pytest.mark.asyncio
async def test_get_resource() -> None:
    group_repository = PrioritySelectedProjectsRepository([
        FakeRepository(),
        FakeRepository(
            resources={
                "numpy.whl": model.HttpResource("url"),
            },
        ),
        FakeRepository(
            resources={
                "numpy.whl": model.HttpResource("wrog"),
            },
        ),
    ])

    resp = await group_repository.get_resource("numpy", "numpy.whl")
    assert resp == model.HttpResource("url")


@pytest.mark.asyncio
async def test_get_resource_failed() -> None:
    group_repository = PrioritySelectedProjectsRepository([
        FakeRepository() for _ in range(3)
    ])

    with pytest.raises(errors.ResourceUnavailable, match="numpy.whl"):
        await group_repository.get_resource("numpy", "numpy.whl")



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import json
import pathlib

import httpx
import pytest
from pytest_httpx import HTTPXMock

from .. import errors, utils


@pytest.mark.parametrize(
        ("url", "url_base", "expected_url"), [
            ("/relative/path", "https://example.com/", "https://example.com/relative/path"),
            ("https://example.com/absolute/path", "https://example.com/", "https://example.com/absolute/path"),
            ("//example.com/path", "https://example.org/", "https://example.com/path"),
            ("http://example.com/path", "https://example.org/", "http://example.com/path"),
        ],
)
def test_url_absolutizer(url: str, url_base: str, expected_url: str) -> None:
    assert utils.url_absolutizer(url, url_base) == expected_url


@pytest.mark.parametrize(
    "json_string", [
        "42", "true", '["a", "b"]', "null", '"ciao"',
    ],
)
@pytest.mark.asyncio
async def test_load_config_json_not_a_dict(
    json_string: str,
    tmp_path: pathlib.PosixPath,
) -> None:
    file = tmp_path / "config.json"
    file.write_text(
        data=json_string,
    )
    with pytest.raises(
        errors.InvalidConfigurationError,
        match=(
            f"Invalid configuration file. {str(file)}"
            " must contain a dictionary."
        ),
    ):
        utils.load_config_json(file)


@pytest.mark.asyncio
async def test_load_config_malformed_json(
    tmp_path: pathlib.PosixPath,
) -> None:
    file = tmp_path / "config.json"
    file.write_text(
        data="{",
    )
    with pytest.raises(
        errors.InvalidConfigurationError,
        match=(
            "Invalid json file"
        ),
    ) as exc_info:
        utils.load_config_json(file)
    assert isinstance(exc_info.value.__cause__, json.JSONDecodeError)


@pytest.mark.asyncio
async def test_load_config_file_not_found(
    tmp_path: pathlib.PosixPath,
) -> None:
    file = tmp_path / "config.json"
    with pytest.raises(
        errors.InvalidConfigurationError,
        match=(
            "Configuration file not found"
        ),
    ) as exc_info:
        utils.load_config_json(file)
    assert isinstance(exc_info.value.__cause__, FileNotFoundError)


@pytest.mark.asyncio
async def test_download_file(tmp_path: pathlib.PosixPath, httpx_mock: HTTPXMock) -> None:
    download_url = "https://example.com/package.tar.gz"
    dest_file = tmp_path / "package.tar.gz"

    httpx_mock.add_response(content="my_file")
    async with httpx.AsyncClient() as http_client:
        await utils.download_file(download_url, dest_file, http_client)

    assert dest_file.exists()
    assert dest_file.read_text() == "my_file"



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

from dataclasses import replace
import fnmatch
import html
import pathlib
import typing

import aiosqlite
from packaging.utils import canonicalize_name

from .. import errors, model
from .. import packaging as _packaging
from .. import utils
from .core import RepositoryContainer, SimpleRepository


class YankProvider(typing.Protocol):
    async def yanked_versions(self, project_page: model.ProjectDetail) -> dict[str, str]:
        ...

    async def yanked_files(self, project_page: model.ProjectDetail) -> dict[str, str]:
        ...


class SqliteYankProvider(YankProvider):
    def __init__(self, database: aiosqlite.Connection) -> None:
        # TODO: Use a synchronization mechanism instead.
        self._initialise_db = True
        self._database = database

    async def _init_db(self) -> None:
        if not self._initialise_db:
            return

        await self._database.execute(
            "CREATE TABLE IF NOT EXISTS yanked_versions"
            "(project_name TEXT, version TEXT, reason TEXT,"
            " date TIMESTAMP DEFAULT CURRENT_TIMESTAMP"
            ", CONSTRAINT pk PRIMARY KEY (project_name, version))",
        )
        await self._database.execute(
            "CREATE TABLE IF NOT EXISTS yanked_releases"
            "(project_name TEXT, file_name TEXT, reason TEXT,"
            " date TIMESTAMP DEFAULT CURRENT_TIMESTAMP"
            ", CONSTRAINT pk PRIMARY KEY (project_name, file_name))",
        )
        self._initialise_db = False

    async def yanked_versions(self, project_page: model.ProjectDetail) -> dict[str, str]:
        await self._init_db()

        query = "SELECT version, reason FROM yanked_versions WHERE project_name = :project_name"
        async with self._database.execute(query, {"project_name": project_page.name}) as cur:
            result = await cur.fetchall()
        return {
            version: reason for version, reason in result
        }

    async def yanked_files(self, project_page: model.ProjectDetail) -> dict[str, str]:
        await self._init_db()

        query = "SELECT file_name, reason FROM yanked_releases WHERE project_name = :project_name"
        async with self._database.execute(query, {"project_name": project_page.name}) as cur:
            result = await cur.fetchall()
        return {
            filename: reason for filename, reason in result
        }


class GlobYankProvider(YankProvider):
    """Yanks distributions according to the provided json configuration file.
    The file MUST contain a dictionary mapping project names to glob patterns
    and yank reasons. For a given project, all files matching the pattern will
    be yanked with the given reason.

    The configuration file must have the following structure:

        {
            "numpy": ["*.exe", "unsupported"],
            "tensorflow": ["*[!.whl]", "temporary"]
        }
    """
    def __init__(
        self,
        yank_config_file: pathlib.Path,
    ) -> None:
        self._yank_config: dict[str, tuple[str, str]] = self._load_config_json(yank_config_file)

    async def yanked_versions(self, project_page: model.ProjectDetail) -> dict[str, str]:
        # TODO: Manage yanked_versions in this component
        return {}

    async def yanked_files(self, project_page: model.ProjectDetail) -> dict[str, str]:
        yanked_files = {}
        if value := self._yank_config.get(project_page.name):
            pattern, reason = value
            yanked_files = {
                file.filename: reason for file in project_page.files
                if fnmatch.fnmatch(file.filename, pattern)
            }

        return yanked_files

    def _load_config_json(self, json_file: pathlib.Path) -> dict[str, tuple[str, str]]:
        json_config = utils.load_config_json(json_file)

        config_dict: dict[str, tuple[str, str]] = {}
        for key, value in json_config.items():
            if (
                not isinstance(key, str) or
                not isinstance(value, list) or
                len(value) != 2 or
                not all(isinstance(elem, str) for elem in value)
            ):
                raise errors.InvalidConfigurationError(
                    f'Invalid yank configuration file. {str(json_file)} must'
                    ' contain a dictionary mapping a project name to a tuple'
                    ' containing a glob pattern and a yank reason.',
                )
            config_dict[canonicalize_name(key)] = (value[0], value[1])

        return config_dict


def update_yanked_attribute(file: model.File, reason: str) -> model.File:
    if reason == '':
        yanked: bool | str = True
    else:
        yanked = html.escape(reason)
    return replace(file, yanked=yanked)


class YankRepository(RepositoryContainer):
    """
    A class that adds support for PEP-592 yank to a SimpleRepository.

    The information related to which version or file of a project are yanked
    comes from a provider that specialized the protocol YankProvider.
    """
    def __init__(
        self,
        source: SimpleRepository,
        yank_provider: YankProvider,
    ) -> None:
        self._yank_provider = yank_provider
        super().__init__(source)

    async def get_project_page(
        self,
        project_name: str,
        *,
        request_context: model.RequestContext = model.RequestContext.DEFAULT,
    ) -> model.ProjectDetail:
        project_page = await super().get_project_page(
            project_name,
            request_context=request_context,
        )

        yanked_versions = await self._yank_provider.yanked_versions(project_page)
        yanked_files = await self._yank_provider.yanked_files(project_page)

        if yanked_versions or yanked_files:
            project_page = self._add_yanked_attribute(
                project_page=project_page,
                yanked_files=yanked_files,
                yanked_versions=yanked_versions,
            )

        return project_page

    def _add_yanked_attribute(
        self,
        project_page: model.ProjectDetail,
        yanked_versions: dict[str, str],
        yanked_files: dict[str, str],
    ) -> model.ProjectDetail:
        files = []
        for file in project_page.files:
            if file.yanked:
                # Skip already yanked files
                pass
            elif reason := yanked_files.get(file.filename):
                file = update_yanked_attribute(file, reason)
            else:
                try:
                    version = _packaging.extract_package_version(
                        filename=file.filename,
                        project_name=canonicalize_name(project_page.name),
                    )
                except ValueError:
                    pass
                else:
                    if reason := yanked_versions.get(version):
                        file = update_yanked_attribute(file, reason)

            files.append(file)

        return replace(project_page, files=tuple(files))



# Copyright (C) 2023, CERN
# This software is distributed under the terms of the MIT
# licence, copied verbatim in the file "LICENSE".
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as Intergovernmental Organization
# or submit itself to any jurisdiction.

import simple_repository


def test_version() -> None:
    assert simple_repository.__version__ != "0.0.0"
